# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from pathlib import Path
from shutil import which
import xml.etree.ElementTree as ET
from codetiming import Timer

from loki.frontend.source import Source
from loki.frontend.util import OMNI, sanitize_ir

from loki import ir
from loki.ir import (
    GenericVisitor, FindNodes, Transformer,
    process_dimension_pragmas, pragmas_attached
)
from loki.expression import (
    symbols as sym, operations as op,
    ExpressionDimensionsMapper, StringConcat, AttachScopesMapper
)
from loki.logging import debug, info, warning, error
from loki.config import config
from loki.tools import (
    as_tuple, execute, gettempdir, filehash, CaseInsensitiveDict
)
from loki.types import BasicType, DerivedType, ProcedureType, SymbolAttributes


__all__ = ['HAVE_OMNI', 'parse_omni_source', 'parse_omni_file', 'parse_omni_ast']


HAVE_OMNI = which('F_Front') is not None
"""Indicate whether OMNI frontend is available."""


@Timer(logger=debug, text=lambda s: f'[Loki::OMNI] Executed parse_omni_file in {s:.2f}s')
def parse_omni_file(filename, xmods=None):
    """
    Deploy the OMNI compiler's frontend (F_Front) to generate the OMNI AST.

    Note that the intermediate XML files can be dumped to file via by setting
    the environment variable ``LOKI_OMNI_DUMP_XML``.
    """
    if not HAVE_OMNI:
        error('OMNI is not available. Is "F_Front" in the search path?')

    dump_xml_files = config['omni-dump-xml']

    filepath = Path(filename)
    info(f'[Loki::OMNI] Parsing {filepath}')

    xml_path = filepath.with_suffix('.xml')
    xmods = xmods or []

    cmd = ['F_Front', '-fleave-comment']
    for m in xmods:
        cmd += ['-M', f'{Path(m)}']
    cmd += [f'{filepath}']

    if dump_xml_files:
        # Parse AST from xml file dumped to disk
        cmd += ['-o', f'{xml_path}']
        execute(cmd)
        return ET.parse(str(xml_path)).getroot()

    result = execute(cmd, silent=False, capture_output=True, text=True)
    return ET.fromstring(result.stdout)


@Timer(logger=debug, text=lambda s: f'[Loki::OMNI] Executed parse_omni_source in {s:.2f}s')
def parse_omni_source(source, filepath=None, xmods=None):
    """
    Deploy the OMNI compiler's frontend (F_Front) to AST for a source string.
    """
    # Use basename of filepath if given
    if filepath is None:
        filepath = Path(filehash(source, prefix='omni-', suffix='.f90'))
    else:
        filepath = filepath.with_suffix(f'.omni{filepath.suffix}')

    # Always store intermediate flies in tmp dir
    filepath = gettempdir()/filepath.name

    debug(f'[Loki::OMNI] Writing temporary source {filepath}')
    with filepath.open('w') as f:
        f.write(source)

    return parse_omni_file(filename=filepath, xmods=xmods)


@Timer(logger=debug, text=lambda s: f'[Loki::OMNI] Executed parse_omni_ast in {s:.2f}s')
def parse_omni_ast(ast, definitions=None, type_map=None, symbol_map=None,
                   raw_source=None, scope=None):
    """
    Generate an internal IR from the raw OMNI parser AST.
    """
    # Parse the raw OMNI language AST
    _ir = OMNI2IR(type_map=type_map, definitions=definitions, symbol_map=symbol_map,
                  raw_source=raw_source, scope=scope).visit(ast)

    # Perform some minor sanitation tasks
    _ir = sanitize_ir(_ir, OMNI)

    return _ir


class OMNI2IR(GenericVisitor):
    # pylint: disable=unused-argument  # Stop warnings about unused arguments

    _omni_types = {
        'Fint': 'INTEGER',
        'Freal': 'REAL',
        'Flogical': 'LOGICAL',
        'Fcharacter': 'CHARACTER',
        'Fcomplex': 'COMPLEX',
        'int': 'INTEGER',
        'real': 'REAL',
    }

    def __init__(self, definitions=None, type_map=None, symbol_map=None,
                 raw_source=None, scope=None):
        super().__init__()

        self.definitions = CaseInsensitiveDict((d.name, d) for d in as_tuple(definitions))
        self.type_map = type_map or {}
        self.symbol_map = symbol_map or {}
        self.raw_source = raw_source.splitlines(keepends=True)
        self.default_scope = scope
        self.lineno = None  # use to save lineno of last element with attribute lineno

    @staticmethod
    def warn_or_fail(msg):
        if config['frontend-strict-mode']:
            error(msg)
            raise NotImplementedError
        warning(msg)

    def type_from_type_attrib(self, type_attrib, **kwargs):
        """
        Helper routine to derive :any:`SymbolAttributes` for a given type name/hash/id
        """
        if type_attrib in self._omni_types:
            typename = self._omni_types[type_attrib]
            _type = SymbolAttributes(BasicType.from_fortran_type(typename))
        elif type_attrib in self.type_map:
            _type = self.visit(self.type_map[type_attrib], **kwargs)
            dims = self.type_map[type_attrib].findall('indexRange')
            if dims:
                dimensions = as_tuple(self.visit(d, **kwargs) for d in dims)
                _type = _type.clone(shape=dimensions)
        else:
            _type = SymbolAttributes(BasicType.from_fortran_type(type_attrib))
        return _type

    def lookup_method(self, instance):
        """
        Alternative lookup method for XML element types, identified by ``element.tag``
        """
        tag = instance.tag.replace('-', '_')
        if tag in self._handlers:
            return self._handlers[tag]
        return super().lookup_method(instance)

    def get_source(self, o):
        """Helper method that builds the source object for a node"""
        file = o.attrib.get('file', None)
        lineno = o.attrib.get('lineno', self.lineno)
        if lineno:
            self.lineno = int(lineno)
            lines = (self.lineno, self.lineno)
            string = self.raw_source[self.lineno-1]
        else:
            lines = (None, None)
            string = None
        return Source(lines=lines, string=string, file=file)

    def visit(self, o, **kwargs):  # pylint: disable=arguments-differ
        """
        Generic dispatch method that tries to generate meta-data from source.
        """
        kwargs['source'] = self.get_source(o)
        kwargs.setdefault('scope', self.default_scope)
        kwargs.setdefault('symbol_map', self.symbol_map)
        return super().visit(o, **kwargs)

    def visit_Element(self, o, **kwargs):
        """
        Universal default for XML element types
        """
        warning('No specific handler for node type %s', o.__class__.name)
        children = tuple(self.visit(c, **kwargs) for c in o)
        children = tuple(c for c in children if c is not None)
        if len(children) == 1:
            return children[0]  # Flatten hierarchy if possible
        return children if len(children) > 0 else None

    def visit_XcodeProgram(self, o, **kwargs):
        body = [self.visit(c, **kwargs) for c in o.find('globalDeclarations')]
        return ir.Section(body=as_tuple(body))

    def visit_FuseDecl(self, o, **kwargs):
        # No ONLY list
        nature = 'intrinsic' if o.attrib.get('intrinsic') == 'true' else None
        name = o.attrib['name']
        scope = kwargs['scope']

        # Rename list
        rename_list = dict(self.visit(s, **kwargs) for s in o.findall('rename'))

        module = self.definitions.get(name, None)
        if module is not None:
            # Import symbol attributes from module, if available
            for k, v in module.symbol_attrs.items():
                # Don't import private module symbols
                if v.private or (module.default_access_spec == "private" and not v.public):
                    continue
                if k in rename_list:
                    local_name = rename_list[k].name
                    scope.symbol_attrs[local_name] = v.clone(imported=True, module=module, use_name=k)
                else:
                    # Need to explicitly reset use_name in case we are importing a symbol
                    # that stems from an import with a rename-list
                    scope.symbol_attrs[k] = v.clone(imported=True, module=module, use_name=None)
        elif rename_list:
            # Module not available but some information via rename-list
            scope.symbol_attrs.update({v.name: v.type.clone(imported=True, use_name=k) for k, v in rename_list.items()})
        rename_list = tuple(rename_list.items()) if rename_list else None
        return ir.Import(module=name, nature=nature, rename_list=rename_list, c_import=False, source=kwargs['source'])

    def visit_FuseOnlyDecl(self, o, **kwargs):
        # ONLY list given (import only selected symbols)
        nature = 'intrinsic' if o.attrib.get('intrinsic') == 'true' else None
        name = o.attrib['name']
        scope = kwargs['scope']
        symbols = tuple(self.visit(c, **kwargs) for c in o.findall('renamable'))
        if nature == 'intrinsic':
            module = None
        else:
            module = self.definitions.get(name, None)

        deferred_type = SymbolAttributes(BasicType.DEFERRED, imported=True)
        if module is None:
            # Initialize symbol attributes as DEFERRED
            for s in symbols:
                if isinstance(s, tuple):  # Renamed symbol
                    scope.symbol_attrs[s[1].name] = deferred_type.clone(use_name=s[0])
                else:
                    scope.symbol_attrs[s.name] = deferred_type
        else:
            # Import symbol attributes from module
            for s in symbols:
                if isinstance(s, tuple):  # Renamed symbol
                    _type = module.symbol_attrs.get(s[0], deferred_type)
                    scope.symbol_attrs[s[1].name] = _type.clone(
                        imported=True, module=module, use_name=s[0]
                    )
                else:
                    # Need to explicitly reset use_name in case we are importing a symbol
                    # that stems from an import with a rename-list
                    _type = module.symbol_attrs.get(s.name, deferred_type)
                    scope.symbol_attrs[s.name] = _type.clone(
                        imported=True, module=module, use_name=None
                    )
        symbols = tuple(
            s[1].rescope(scope=scope) if isinstance(s, tuple) else s.rescope(scope=scope) for s in symbols
        )
        return ir.Import(module=name, symbols=symbols, nature=nature, c_import=False, source=kwargs['source'])

    def visit_renamable(self, o, **kwargs):
        name = o.attrib['use_name']
        if o.attrib.get('is_operator') == 'true':
            if name == '=':
                name = 'ASSIGNMENT(=)'
            else:
                name = f'OPERATOR({name})'

        if o.attrib.get('local_name'):
            return (name, sym.Variable(name=o.attrib['local_name']))
        return sym.Variable(name=name)

    visit_rename = visit_renamable

    def visit_FinterfaceDecl(self, o, **kwargs):
        abstract = o.get('is_abstract') == 'true'

        if o.get('is_assignment') == 'true':
            name = 'ASSIGNMENT(=)'
        elif o.get('is_operator') == 'true':
            name = f'OPERATOR({o.get("name")})'
        else:
            name = o.get('name')

        if name is not None:
            scope = kwargs['scope']
            if name not in scope.symbol_attrs:
                scope.symbol_attrs[name] = SymbolAttributes(ProcedureType(name, is_generic=True))
            spec = sym.Variable(name=name, scope=kwargs['scope'])
        else:
            spec = None

        body = tuple(self.visit(c, **kwargs) for c in o)
        return ir.Interface(body=body, abstract=abstract, spec=spec, source=kwargs['source'])

    def _create_Subroutine_object(self, o, scope, symbol_map):
        """Helper method to instantiate a Subroutine object"""
        from loki.subroutine import Subroutine  # pylint: disable=import-outside-toplevel,cyclic-import
        assert o.tag in ('FfunctionDefinition', 'FfunctionDecl')
        name = o.find('name').text

        # Check if the Subroutine node has been created before by looking it up in the scope
        routine = None
        if scope is not None and name in scope.symbol_attrs:
            proc_type = scope.symbol_attrs[name]  # Look-up only in current scope!
            if proc_type and proc_type.dtype.procedure != BasicType.DEFERRED:
                routine = proc_type.dtype.procedure
                if not routine._incomplete:
                    # We return the existing object right away, unless it exists from a
                    # previous incomplete parse for which we have to make sure we get a
                    # full parse first
                    return routine

        # Return type and dummy args
        ftype = self.type_map[o.find('name').attrib['type']]
        if ftype.attrib.get('is_program') == 'true':
            self.warn_or_fail('No support for PROGRAM')
            return None
        proc_type = self.visit(ftype, scope=scope, symbol_map=symbol_map)
        is_function = ftype.attrib['return_type'] != 'Fvoid'
        args = tuple(a.text for a in ftype.findall('params/name'))

        # Function/Subroutine prefix
        prefix = proc_type.prefix or ()
        if prefix:
            # We store the prefix on the Subroutine object, so let's remove it from the symbol attrs
            proc_type = proc_type.clone(prefix=None)

        # Function suffix (result name and language binding, but no support for the latter in OMNI)
        result = ftype.attrib.get('result_name')

        # Instantiate the object
        if routine is None:
            routine = Subroutine(
                name=name, args=args, prefix=prefix, bind=None,
                result_name=result, is_function=is_function, parent=scope,
                ast=o, source=self.get_source(o)
            )
        else:
            routine.__initialize__(
                name=name, args=args, docstring=routine.docstring, spec=routine.spec,
                body=routine.body, contains=routine.contains, prefix=prefix, bind=None,
                result_name=result, is_function=is_function, ast=o,
                source=self.get_source(o), incomplete=routine._incomplete
            )

        return routine

    def visit_FfunctionDefinition(self, o, **kwargs):
        # Update the symbol map with local entries
        kwargs['symbol_map'] = kwargs['symbol_map'].copy()
        kwargs['symbol_map'].update({s.attrib['type']: s for s in o.find('symbols')})

        # Instantiate the object
        routine = self._create_Subroutine_object(o, kwargs['scope'], kwargs['symbol_map'])
        if routine is None:
            return None
        kwargs['scope'] = routine

        # Parse the spec
        spec = self.visit(o.find('declarations'), **kwargs)
        spec = sanitize_ir(spec, OMNI)

        # Filter out the declaration for the subroutine name but keep it for functions (since
        # this declares the return type)
        spec_map = {}
        if not routine.is_function:
            spec_map.update({
                d: None for d in FindNodes((ir.ProcedureDeclaration, ir.VariableDeclaration)).visit(spec)
                if routine.name in d.symbols
            })

        # Hack: We remove comments from the beginning of the spec to get the docstring
        docstring = []
        for node in spec.body:
            if node in spec_map:
                continue
            if not isinstance(node, (ir.Comment, ir.CommentBlock)):
                break
            docstring.append(node)
            spec_map[node] = None
        docstring = as_tuple(docstring)
        spec = Transformer(spec_map, invalidate_source=False).visit(spec)

        # Insert the `implicit none` statement OMNI omits (slightly hacky!)
        f_imports = [im for im in FindNodes(ir.Import).visit(spec) if not im.c_import]
        if not f_imports:
            spec.prepend(ir.Intrinsic(text='IMPLICIT NONE'))
        else:
            spec.insert(spec.body.index(f_imports[-1])+1, ir.Intrinsic(text='IMPLICIT NONE'))

        # Parse member functions
        body_ast = o.find('body')
        contains_ast = None if body_ast is None else body_ast.find('FcontainsStatement')
        if contains_ast is not None:
            contains = self.visit(contains_ast, **kwargs)

            # Strip contains part from the XML before we proceed
            body_ast.remove(contains_ast)
        else:
            contains = None

        # Finally, take care of the body
        if body_ast is None:
            body = ir.Section(body=())
        else:
            body = ir.Section(body=self.visit(body_ast, **kwargs))
            body = sanitize_ir(body, OMNI)

        # Finally, call the subroutine constructor on the object again to register all
        # bits and pieces in place and rescope all symbols
        # pylint: disable=unnecessary-dunder-call
        routine.__initialize__(
            name=routine.name, args=routine._dummies, docstring=docstring, spec=spec,
            body=body, contains=contains, ast=o, prefix=routine.prefix,
            bind=routine.bind, result_name=routine.result_name,
            is_function=routine.is_function, rescope_symbols=True,
            source=routine.source, incomplete=False
        )

        # For deferred array dimensions on allocatables, we infer the conceptual
        # dimension by finding any `allocate(var(<dims>))` statements.
        routine._infer_allocatable_shapes()

        # Update array shapes with Loki dimension pragmas
        with pragmas_attached(routine, ir.VariableDeclaration):
            routine.spec = process_dimension_pragmas(routine.spec, scope=routine)

        return routine

    visit_FfunctionDecl = visit_FfunctionDefinition

    def visit_FcontainsStatement(self, o, **kwargs):
        body = [self.visit(c, **kwargs) for c in o]
        body = [c for c in body if c is not None]
        body = [ir.Intrinsic('CONTAINS', source=kwargs['source'])] + body
        return ir.Section(body=as_tuple(body))

    def visit_FmoduleProcedureDecl(self, o, **kwargs):
        symbols = as_tuple(self.visit(o.find('name'), **kwargs))
        symbols = AttachScopesMapper()(symbols, scope=kwargs['scope'])
        return ir.ProcedureDeclaration(symbols=symbols, module=True, source=kwargs.get('source'))

    def _create_Module_object(self, o, scope):
        """Helper method to instantiate a Module object"""
        from loki.module import Module  # pylint: disable=import-outside-toplevel,cyclic-import

        name = o.attrib['name']

        # Check if the Module node has been created before by looking it up in the scope
        if scope is not None and name in scope.symbol_attrs:
            module_type = scope.symbol_attrs[name]  # Look-up only in current scope
            if module_type and module_type.dtype.module != BasicType.DEFERRED:
                return module_type.dtype.module

        module = Module(name=name, parent=scope)
        self.definitions[name] = module
        return module

    def visit_FmoduleDefinition(self, o, **kwargs):
        # Update the symbol map with local entries
        kwargs['symbol_map'] = kwargs['symbol_map'].copy()
        kwargs['symbol_map'].update({s.attrib['type']: s for s in o.find('symbols')})

        # Instantiate the object
        module = self._create_Module_object(o, kwargs['scope'])
        kwargs['scope'] = module

        # Pre-populate symbol table with procedure types declared in this module
        # to correctly classify inline function calls and type-bound procedures
        contains_ast = o.find('FcontainsStatement')
        if contains_ast is not None:
            # Note that we overwrite this variable subsequently with the fully parsed subroutines
            # where the visit-method for the subroutine/function statement will pick out the existing
            # subroutine objects using the weakref pointers stored in the symbol table.
            # I know, it's not pretty but alternatively we could hand down this array as part of
            # kwargs but that feels like carrying around a lot of bulk, too.
            contains = [
                self._create_Subroutine_object(member_ast, kwargs['scope'], kwargs['symbol_map'])
                for member_ast in contains_ast.findall('FfunctionDefinition')
            ]

        # Parse the spec
        spec = self.visit(o.find('declarations'), **kwargs)
        spec = sanitize_ir(spec, OMNI)

        # Hack: We remove comments from the beginning of the spec to get the docstring
        docstring = []
        spec_map = {}
        for node in spec.body:
            if node in spec_map:
                continue
            if not isinstance(node, (ir.Comment, ir.CommentBlock)):
                break
            docstring.append(node)
            spec_map[node] = None
        docstring = as_tuple(docstring)
        spec = Transformer(spec_map, invalidate_source=False).visit(spec)

        # Parse member functions
        if contains_ast is not None:
            contains = self.visit(contains_ast, **kwargs)
        else:
            contains = None

        # Finally, call the module constructor on the object again to register all
        # bits and pieces in place and rescope all symbols
        # pylint: disable=unnecessary-dunder-call
        module.__initialize__(
            name=module.name, docstring=docstring, spec=spec, contains=contains,
            ast=o, rescope_symbols=True, source=kwargs['source'], incomplete=False
        )

        return module

    def visit_declarations(self, o, **kwargs):
        body = tuple(self.visit(c, **kwargs) for c in o)
        body = tuple(c for c in body if c is not None)
        return ir.Section(body=body, source=kwargs['source'])

    def visit_body(self, o, **kwargs):
        body = tuple(self.visit(c, **kwargs) for c in o)
        body = tuple(c for c in body if c is not None)
        return body

    def visit_FimportDecl(self, o, **kwargs):
        symbols = tuple(self.visit(i, **kwargs) for i in o)
        symbols = AttachScopesMapper()(symbols, scope=kwargs['scope'])
        return ir.Import(
            module=None, symbols=symbols, f_import=True, source=kwargs['source']
        )

    def visit_varDecl(self, o, **kwargs):
        # OMNI has only one variable per declaration, find and create that
        name = o.find('name')
        variable = self.visit(name, **kwargs)

        interface = None
        scope = kwargs['scope']

        # Create the declared type
        if name.attrib['type'] in self._omni_types:
            # Intrinsic scalar type
            t = self._omni_types[name.attrib['type']]
            _type = SymbolAttributes(BasicType.from_fortran_type(t))
            dimensions = None

        elif name.attrib['type'] in self.type_map:
            # Type with attributes or derived type
            tast = self.type_map[name.attrib['type']]
            _type = self.visit(tast, **kwargs)

            dimensions = as_tuple(self.visit(d, **kwargs) for d in tast.findall('indexRange'))
            if dimensions:
                _type = _type.clone(shape=dimensions)
                variable = variable.clone(dimensions=dimensions)

            if isinstance(_type.dtype, ProcedureType):
                if _type.dtype.name == 'UNKNOWN':
                    # _Probably_ a declaration with implicit interface
                    dtype = ProcedureType(
                        variable.name, is_function=_type.dtype.is_function, return_type=_type.dtype.return_type
                    )
                    _type = _type.clone(dtype=dtype)
                    interface = dtype.return_type.dtype

                if variable != scope.name:
                    # Instantiate the symbol representing the procedure in the current scope to create
                    # relevant symbol table entries, and then extract the dtype
                    try:
                        symbol_scope = scope.get_symbol_scope(_type.dtype.name)
                        interface = symbol_scope.Variable(name=_type.dtype.name)
                        _type = _type.clone(dtype=interface.type.dtype)
                    except AttributeError:
                        # Interface symbol could not be found
                        pass

                elif _type.dtype.return_type is not None:
                    # This is the declaration of the return type inside a function, which is
                    # why we restore the return_type
                    _type = _type.dtype.return_type

                    # If the return type has a shape, we need to apply this as a dimension to the
                    # variable, otherwise it will be missing from the declaration
                    if _type.shape:
                        variable = variable.clone(dimensions=_type.shape)

                if tast.attrib.get('is_external') == 'true':
                    _type.external = True

        else:
            raise ValueError

        if o.find('value') is not None:
            _type = _type.clone(initial=AttachScopesMapper()(self.visit(o.find('value'), **kwargs), scope=scope))
        if _type.kind is not None:
            _type = _type.clone(kind=AttachScopesMapper()(_type.kind, scope=scope))

        scope.symbol_attrs[variable.name] = _type
        variable = variable.rescope(scope=scope)

        if isinstance(_type.dtype, ProcedureType):
            # This is actually a function or subroutine (EXTERNAL or PROCEDURE declaration)
            return ir.ProcedureDeclaration(
                symbols=(variable,), interface=interface, external=_type.external or False,
                source=kwargs['source']
            )

        return ir.VariableDeclaration(symbols=(variable,), source=kwargs['source'])

    def visit_FstructDecl(self, o, **kwargs):
        name = o.find('name')
        struct_type = self.type_map[name.attrib['type']]

        # Type attributes
        abstract = struct_type.get('is_abstract') == 'true'
        if 'extends' in struct_type.attrib:
            base_type = kwargs['symbol_map'][struct_type.attrib['extends']]
            extends = base_type.find('name').text
        else:
            extends = None
        bind_c = struct_type.get('bind', '').lower() == 'c'
        private = struct_type.get('is_private', '').lower() == 'true'
        public = struct_type.get('is_public', '').lower() == 'true'

        # Type Parameters
        if struct_type.find('typeParams') is not None:
            self.warn_or_fail('Parameterized types not implemented')

        # Instantiate the TypeDef without its body
        # Note: This creates the symbol table for the declarations and
        # the typedef object registers itself in the parent scope
        typedef = ir.TypeDef(
            name=name.text, body=(), abstract=abstract, extends=extends, bind_c=bind_c,
            private=private, public=public, parent=kwargs['scope'], source=kwargs['source']
        )
        kwargs['scope'] = typedef

        body = []

        # Check if the type is marked as sequence
        if struct_type.get('is_sequence') == 'true':
            body += [ir.Intrinsic('SEQUENCE')]

        # Build the list of derived type members and individual body for each
        if struct_type.find('symbols'):
            variables = self.visit(struct_type.find('symbols'), **kwargs)
            for v in variables:
                if isinstance(v.type.dtype, ProcedureType):
                    if v.type.dtype.name == v and v.type.dtype.is_function:
                        interface = v.type.dtype.return_type
                    else:
                        iface_name = v.type.dtype.name
                        interface = sym.Variable(name=iface_name, scope=kwargs['scope'].get_symbol_scope(iface_name))
                    body += [ir.ProcedureDeclaration(symbols=(v,), interface=interface)]
                else:
                    body += [ir.VariableDeclaration(symbols=(v,))]

        if struct_type.find('typeBoundProcedures'):
            # See if components are marked private
            body += [ir.Intrinsic('CONTAINS')]
            if struct_type.attrib.get('is_internal_private') == 'true':
                body += [ir.Intrinsic('PRIVATE')]
            body += self.visit(struct_type.find('typeBoundProcedures'), **kwargs)

        # Finally: update the typedef with its body
        typedef._update(body=as_tuple(body))
        typedef.rescope_symbols()
        return typedef

    def visit_symbols(self, o, **kwargs):
        """
        Build the list of variables for a `FstructType` node
        """
        variables = []
        for s in o:
            var = self.visit(s.find('name'), **kwargs)
            _type = self.type_from_type_attrib(s.attrib['type'], **kwargs)
            kwargs['scope'].symbol_attrs[var.name] = _type

            if _type.shape:
                var = var.clone(dimensions=_type.shape)
            variables += [var.rescope(scope=kwargs['scope'])]
        return variables

    def visit_typeBoundProcedures(self, o, **kwargs):
        procedures = []
        for i in o:
            proc = self.visit(i, **kwargs)
            if i.get('is_deferred') == 'true':
                assert proc.type.deferred is True
                assert proc.type.bind_names and len(proc.type.bind_names) == 1
                intf = proc.type.bind_names[0]
                procedures += [ir.ProcedureDeclaration(interface=intf, symbols=(proc,))]
            elif i.tag == 'typeBoundGenericProcedure':
                procedures += [ir.ProcedureDeclaration(symbols=(proc,), generic=True)]
            elif i.tag == 'finalProcedure':
                procedures += [ir.ProcedureDeclaration(symbols=(proc,), final=True)]
            else:
                procedures += [ir.ProcedureDeclaration(symbols=(proc,))]
        return procedures

    def visit_typeBoundProcedure(self, o, **kwargs):
        scope = kwargs['scope']
        var = self.visit(o.find('name'), **kwargs)

        _type = self.type_from_type_attrib(o.attrib['type'], **kwargs)
        if o.get('pass') == 'pass':
            _type = _type.clone(pass_attr=o.get('pass_arg_name', True))
        elif o.get('pass') == 'nopass':
            _type = _type.clone(pass_attr=False)
        if o.get('is_deferred') == 'true':
            _type = _type.clone(deferred=True)
        if o.get('is_non_overridable') == 'true':
            _type = _type.clone(non_overridable=True)
        if o.get('is_private') == 'true':
            _type = _type.clone(private=True)
        if o.get('is_public') == 'true':
            _type = _type.clone(public=True)

        if o.find('binding'):
            bind_name = self.visit(o.find('binding/name'), **kwargs)
            bind_name_scope = scope.get_symbol_scope(bind_name.name)

            # Set correct type for interface/binding
            if bind_name_scope is not None:
                bind_name = bind_name.rescope(scope=bind_name_scope)
            else:
                bind_name = bind_name.clone(type=bind_name.type.clone(dtype=ProcedureType(bind_name.name)))

            if bind_name.name.lower() == var.name.lower() and not _type.deferred:
                # No need to assign bind_names property
                _type = _type.clone(dtype=bind_name.type.dtype)
            else:
                # Assign the binding as bind_nameial (and park the interface here for
                # declarations with deferred attribute)
                _type = _type.clone(dtype=bind_name.type.dtype, bind_names=(bind_name,))

        scope.symbol_attrs[var.name] = _type
        return var.rescope(scope=scope)

    def visit_typeBoundGenericProcedure(self, o, **kwargs):
        scope = kwargs['scope']
        var = self.visit(o.find('name'), **kwargs)

        _type = SymbolAttributes(ProcedureType(name=var.name, is_generic=True))
        if o.get('is_private') == 'true':
            _type = _type.clone(private=True)
        if o.get('is_public') == 'true':
            _type = _type.clone(public=True)

        assert o.find('binding') is not None

        bind_names = []
        for name in o.findall('binding/name'):
            bind_name = self.visit(name, **kwargs)
            bind_name_scope = scope.get_symbol_scope(bind_name.name)

            # Set correct type for interface/binding
            if bind_name_scope is not None:
                bind_name = bind_name.rescope(scope=bind_name_scope)
            else:
                bind_name = bind_name.clone(type=bind_name.type.clone(dtype=ProcedureType(bind_name.name)))

            bind_names += [bind_name]

        _type = _type.clone(bind_names=as_tuple(bind_names))
        scope.symbol_attrs[var.name] = _type
        return var.rescope(scope=scope)

    def visit_finalProcedure(self, o, **kwargs):
        scope = kwargs['scope']
        var = self.visit(o.find('name'), **kwargs)
        _type = scope.symbol_attrs.lookup(var.name)
        scope.symbol_attrs[var.name] = _type
        return var.rescope(scope=scope)

    def visit_FdataDecl(self, o, **kwargs):
        variable = self.visit(o.find('varList'), **kwargs)
        values = self.visit(o.find('valueList'), **kwargs)
        return ir.DataDeclaration(variable=variable, values=values, source=kwargs['source'])

    def visit_varList(self, o, **kwargs):
        children = tuple(self.visit(c, **kwargs) for c in o)
        children = tuple(c for c in children if c is not None)
        return children

    visit_valueList = visit_varList

    def visit_FbasicType(self, o, **kwargs):
        ref = o.attrib.get('ref', None)
        if ref in self._omni_types:
            dtype = BasicType.from_fortran_type(self._omni_types[ref])
            kind = self.visit(o.find('kind'), **kwargs) if o.find('kind') is not None else None
            length = o.find('len')
            if length is not None:
                if length == '*':
                    pass
                elif length.attrib.get('is_assumed_size') == 'true':
                    length = '*'
                elif length.attrib.get('is_assumed_shape') == 'true':
                    length = ':'
                else:
                    length = self.visit(length, **kwargs)
            _type = SymbolAttributes(dtype, kind=kind, length=length)
        elif ref in self.type_map:
            if o.find('name') is not None:
                _type = self.visit(self.type_map[ref], name=o.find('name').text, **kwargs)
            else:
                _type = self.visit(self.type_map[ref], **kwargs)
            if o.attrib.get('is_class') == 'true':
                _type = _type.clone(polymorphic=True)
        elif ref == 'FnumericAll':
            _type = SymbolAttributes(BasicType.DEFERRED)
        else:
            raise ValueError

        shape = o.findall('indexRange')
        if shape:
            _type.shape = tuple(self.visit(s, **kwargs) for s in shape)

        # OMNI types are build recursively from references (Matroshka-style)
        if o.get('intent') is not None:
            _type.intent = o.get('intent')
        if o.get('is_allocatable') == 'true':
            _type.allocatable = True
        if o.get('is_pointer') == 'true':
            _type.pointer = True
        if o.get('is_optional') == 'true':
            _type.optional = True
        if o.get('is_parameter') == 'true':
            _type.parameter = True
        if o.get('is_target') == 'true':
            _type.target = True
        if o.get('is_contiguous') == 'true':
            _type.contiguous = True
        if o.get('is_private') == 'true':
            _type.private = True
        if o.get('is_public') == 'true':
            _type.public = True
        if o.get('is_save') == 'true':
            _type.save = True
        return _type

    def visit_FfunctionType(self, o, **kwargs):
        if o.attrib['return_type'] == 'Fvoid':
            return_type = None
        elif o.attrib['return_type'] in self._omni_types:
            return_type = SymbolAttributes(BasicType.from_fortran_type(self._omni_types[o.attrib['return_type']]))
        elif o.attrib['return_type'] in self.type_map:
            return_type = self.visit(self.type_map[o.attrib['return_type']], **kwargs)
        else:
            raise ValueError

        if o.attrib['type'] in kwargs['symbol_map']:
            name = kwargs['symbol_map'][o.attrib['type']].find('name').text
        else:
            name = kwargs.get('name', 'UNKNOWN')
        dtype = ProcedureType(name, is_function=return_type is not None, return_type=return_type)

        prefix = []
        if o.attrib.get('is_pure') == 'true':
            prefix += ['PURE']
        if o.attrib.get('is_elemental') == 'true':
            prefix += ['ELEMENTAL']
        if o.attrib.get('is_recursive') == 'true':
            prefix += ['RECURSIVE']
        return SymbolAttributes(dtype, prefix=prefix or None)

    def visit_FstructType(self, o, **kwargs):
        # We have encountered a derived type as part of the declaration in the spec
        # of a routine.
        name = o.attrib['type']
        if name in kwargs['symbol_map']:
            name = kwargs['symbol_map'][name].find('name').text

        # Check if we know that type already
        dtype = kwargs['scope'].symbol_attrs.lookup(name, recursive=True)
        if dtype is None or dtype.dtype == BasicType.DEFERRED:
            dtype = DerivedType(name=name, typedef=BasicType.DEFERRED)
        else:
            dtype = dtype.dtype

        return SymbolAttributes(dtype)

    def visit_value(self, o, **kwargs):
        return self.visit(o[0], **kwargs)

    visit_kind = visit_value
    visit_len = visit_value

    def visit_associateStatement(self, o, **kwargs):
        associations = tuple(self.visit(c, **kwargs) for c in o.findall('symbols/id'))

        # Create a scope for the associate
        parent_scope = kwargs['scope']
        associate = ir.Associate(associations=(), body=(), parent=parent_scope, source=kwargs['source'])
        kwargs['scope'] = associate

        # Put associate expressions into the right scope and determine type of new symbols
        rescoped_associations = []
        for expr, name in associations:
            # Put symbols in associated expression into the right scope
            expr = AttachScopesMapper()(expr, scope=parent_scope)

            # Determine type of new names
            if isinstance(expr, (sym.TypedSymbol, sym.MetaSymbol)):
                # Use the type of the associated variable
                _type = expr.type.clone(parent=None)
                if isinstance(expr, sym.Array) and expr.dimensions is not None:
                    shape = ExpressionDimensionsMapper()(expr)
                    if shape == (sym.IntLiteral(1),):
                        # For a scalar expression, we remove the shape
                        shape = None
                    _type = _type.clone(shape=shape)
            else:
                # TODO: Handle data type and shape of complex expressions
                shape = ExpressionDimensionsMapper()(expr)
                if shape == (sym.IntLiteral(1),):
                    # For a scalar expression, we remove the shape
                    shape = None
                _type = SymbolAttributes(BasicType.DEFERRED, shape=shape)
            name = name.clone(scope=associate, type=_type)
            rescoped_associations += [(expr, name)]
        associations = as_tuple(rescoped_associations)

        body = self.visit(o.find('body'), **kwargs)
        associate._update(associations=associations, body=body)
        return associate

    def visit_id(self, o, **kwargs):
        expr = self.visit(o.find('value'), **kwargs)
        name = self.visit(o.find('name'), **kwargs)
        return expr, name

    def visit_exprStatement(self, o, **kwargs):
        return self.visit(o[0], **kwargs)

    def visit_FcommentLine(self, o, **kwargs):
        return ir.Comment(text=o.text, source=kwargs['source'])

    def visit_FpragmaStatement(self, o, **kwargs):
        keyword = o.text.split(' ')[0]
        content = ' '.join(o.text.split(' ')[1:])
        return ir.Pragma(keyword=keyword, content=content, source=kwargs['source'])

    def visit_FassignStatement(self, o, **kwargs):
        lhs = self.visit(o[0], **kwargs)
        rhs = self.visit(o[1], **kwargs)
        return ir.Assignment(lhs=lhs, rhs=rhs, source=kwargs['source'])

    def visit_FallocateStatement(self, o, **kwargs):
        variables = tuple(self.visit(c, **kwargs) for c in o.findall('alloc'))

        alloc_opts = {}
        if o.find('allocOpt') is not None:
            alloc_opts = [self.visit(opt, **kwargs) for opt in o.findall('allocOpt')]
            alloc_opts = [opt for opt in alloc_opts if opt is not None]
            alloc_opts = dict(alloc_opts)

        return ir.Allocation(variables=variables, source=kwargs['source'],
                             data_source=alloc_opts.get('source'), status_var=alloc_opts.get('stat'))

    def visit_allocOpt(self, o, **kwargs):
        keyword = o.attrib['kind'].lower()
        if keyword in ('source', 'stat'):
            return keyword, self.visit(o[0], **kwargs)
        self.warn_or_fail(f'Unsupported allocation option: {keyword}')
        return None

    def visit_FdeallocateStatement(self, o, **kwargs):
        variables = tuple(self.visit(c, **kwargs) for c in o.findall('alloc'))

        alloc_opts = {}
        if o.find('allocOpt') is not None:
            alloc_opts = [self.visit(opt, **kwargs) for opt in o.findall('allocOpt')]
            alloc_opts = [opt for opt in alloc_opts if opt is not None]
            alloc_opts = dict(alloc_opts)

        return ir.Deallocation(variables=variables, source=kwargs['source'],
                               status_var=alloc_opts.get('stat'))

    def visit_FnullifyStatement(self, o, **kwargs):
        variables = tuple(self.visit(c, **kwargs) for c in o.findall('alloc'))
        return ir.Nullify(variables=variables, source=kwargs['source'])

    def visit_alloc(self, o, **kwargs):
        variable = self.visit(o[0], **kwargs)
        if o.find('arrayIndex') is not None:
            dimensions = tuple(self.visit(c, **kwargs) for c in o.findall('arrayIndex'))
            variable = variable.clone(dimensions=dimensions)
        return variable

    def visit_FwhereStatement(self, o, **kwargs):
        conditions = tuple(self.visit(c, **kwargs) for c in o.findall('condition'))
        bodies = tuple(self.visit(b, **kwargs) for b in o.findall('then/body'))
        if o.find('else') is not None:
            default = self.visit(o.find('else/body'), **kwargs)
        else:
            default = ()
        return ir.MaskedStatement(conditions=conditions, bodies=bodies, default=default, source=kwargs['source'])

    def visit_blockStatement(self, o, **kwargs):
        if (forall_stmt := o.find('body/forallStatement')) is not None:
            return self.visit(forall_stmt, **kwargs)
        self.warn_or_fail('Unsupported blockStatement')
        return None

    def visit_forallStatement(self, o, **kwargs):
        body = self.visit(o.find('body'), **kwargs)
        named_bounds = ()
        for var, index_range in zip(o.findall('Var'), o.findall('indexRange')):
            variable = self.visit(var, **kwargs)
            lower = self.visit(index_range.find('lowerBound'), **kwargs)
            upper = self.visit(index_range.find('upperBound'), **kwargs)
            bounds = sym.RangeIndex((lower, upper))
            named_bounds += ((variable, bounds),)
        if (condition := o.find('condition')) is not None:
            mask = self.visit(condition, **kwargs)
        else:
            mask = None
        return ir.Forall(name=None, named_bounds=named_bounds, body=body,
                         mask=mask, inline=False, source=kwargs.get('source'))

    def visit_FpointerAssignStatement(self, o, **kwargs):
        target = self.visit(o[0], **kwargs)
        expr = self.visit(o[1], **kwargs)
        return ir.Assignment(lhs=target, rhs=expr, ptr=True, source=kwargs['source'])

    def visit_FdoWhileStatement(self, o, **kwargs):
        assert o.find('condition') is not None
        assert o.find('body') is not None
        condition = self.visit(o.find('condition'), **kwargs)
        body = self.visit(o.find('body'), **kwargs)
        return ir.WhileLoop(condition=condition, body=body, source=kwargs['source'])

    def visit_FdoStatement(self, o, **kwargs):
        assert o.find('body') is not None
        body = self.visit(o.find('body'), **kwargs)
        if o.find('Var') is None:
            # We are in an unbound do loop
            return ir.WhileLoop(condition=None, body=body, source=kwargs['source'])
        variable = self.visit(o.find('Var'), **kwargs)
        lower = self.visit(o.find('indexRange/lowerBound'), **kwargs)
        upper = self.visit(o.find('indexRange/upperBound'), **kwargs)
        step = self.visit(o.find('indexRange/step'), **kwargs)
        # Drop OMNI's `:1` step counting for ranges in the name of consistency
        step = None if step == '1' else step
        bounds = sym.LoopRange((lower, upper, step))
        return ir.Loop(variable=variable, body=body, bounds=bounds, source=kwargs['source'])

    def visit_FdoLoop(self, o, **kwargs):
        variable = self.visit(o.find('Var'), **kwargs)
        lower = self.visit(o.find('indexRange/lowerBound'), **kwargs)
        upper = self.visit(o.find('indexRange/upperBound'), **kwargs)
        step = self.visit(o.find('indexRange/step'), **kwargs)
        # Drop OMNI's `:1` step counting for ranges in the name of consistency
        step = None if step == '1' else step
        bounds = sym.LoopRange((lower, upper, step))
        values = as_tuple(self.visit(o.find('value'), **kwargs))
        return sym.InlineDo(values, variable, bounds)

    def visit_FifStatement(self, o, **kwargs):
        condition = self.visit(o.find('condition'), **kwargs)
        body = self.visit(o.find('then/body'), **kwargs)
        if o.find('else'):
            else_body = self.visit(o.find('else/body'), **kwargs)
        else:
            else_body = ()
        return ir.Conditional(condition=condition, body=body, else_body=else_body, source=kwargs['source'])

    def visit_condition(self, o, **kwargs):
        return self.visit(o[0], **kwargs)

    def visit_FselectCaseStatement(self, o, **kwargs):
        expr = self.visit(o.find('value'), **kwargs)
        cases = tuple(self.visit(case, **kwargs) for case in o.findall('FcaseLabel'))
        values, bodies = zip(*cases)
        if None in values:
            else_index = values.index(None)
            else_body = as_tuple(bodies[else_index])
            values = values[:else_index] + values[else_index+1:]
            bodies = bodies[:else_index] + bodies[else_index+1:]
        else:
            else_body = ()

        # Retain comments before the first case
        value_idx, case_idx = list(o).index(o.find('value')), list(o).index(o.find('FcaseLabel'))
        pre = as_tuple(self.visit(c, **kwargs) for c in o[value_idx+1:case_idx])

        return (
            *pre,
            ir.MultiConditional(expr=expr, values=values, bodies=bodies, else_body=else_body,
                                source=kwargs['source'])
        )

    def visit_FcaseLabel(self, o, **kwargs):
        values = [self.visit(value, **kwargs) for value in list(o) if value.tag in ('value', 'indexRange')]
        if not values:
            values = None
        elif len(values) == 1:
            values = values.pop()
        body = self.visit(o.find('body'), **kwargs)
        return as_tuple(values) or None, as_tuple(body)

    def visit_FenumDecl(self, o, **kwargs):
        enum_type = self.type_map[o.attrib['type']]

        # Build the list of symbols
        symbols = []
        for i in enum_type.findall('symbols/id'):
            var = self.visit(i.find('name'), **kwargs)
            initial = i.find('value')
            if initial is not None:
                initial = self.visit(initial, **kwargs)
            _type = SymbolAttributes(BasicType.INTEGER, initial=initial)
            symbols += [var.clone(type=_type)]

        # Put symbols in the right scope (that should register their type in that scope's symbol table)
        symbols = tuple(s.rescope(scope=kwargs['scope']) for s in symbols)

        # Create the enum
        return ir.Enumeration(symbols=symbols, source=kwargs['source'])

    def visit_FmemberRef(self, o, **kwargs):
        parent = self.visit(o.find('varRef'), **kwargs)
        name = f'{parent.name}%{o.attrib["member"]}'
        variable = sym.Variable(name=name, parent=parent)
        return variable

    def visit_name(self, o, **kwargs):
        return sym.Variable(name=o.text)

    visit_Var = visit_name

    def visit_FarrayRef(self, o, **kwargs):
        var = self.visit(o.find('varRef'), **kwargs)
        dimensions = as_tuple(self.visit(i, **kwargs) for i in o[1:])
        var = var.clone(dimensions=dimensions)
        return var

    def visit_varRef(self, o, **kwargs):
        return self.visit(o[0], **kwargs)

    def visit_arrayIndex(self, o, **kwargs):
        return self.visit(o[0], **kwargs)

    def visit_indexRange(self, o, **kwargs):
        lbound = o.find('lowerBound')
        lower = self.visit(lbound, **kwargs) if lbound is not None else None
        ubound = o.find('upperBound')
        upper = self.visit(ubound, **kwargs) if ubound is not None else None
        st = o.find('step')
        step = self.visit(st, **kwargs) if st is not None else None
        # Drop OMNI's `:1` step counting for ranges in the name of consistency
        step = None if step == '1' else step
        return sym.RangeIndex((lower, upper, step))

    def visit_FcharacterRef(self, o, **kwargs):
        var = self.visit(o.find('varRef'), **kwargs)
        dimensions = self.visit(o.find('indexRange'), **kwargs)
        return sym.StringSubscript(var, dimensions)

    def visit_lowerBound(self, o, **kwargs):
        return self.visit(o[0], **kwargs)

    visit_upperBound = visit_lowerBound
    visit_step = visit_lowerBound

    def visit_FrealConstant(self, o, **kwargs):
        if 'kind' in o.attrib and not 'd' in o.text.lower():
            _type = self.visit(self.type_map[o.attrib.get('type')], **kwargs)
            return sym.Literal(value=o.text, type=BasicType.REAL, kind=_type.kind)
        return sym.Literal(value=o.text, type=BasicType.REAL)

    def visit_FlogicalConstant(self, o, **kwargs):
        return sym.Literal(value=o.text, type=BasicType.LOGICAL)

    def visit_FcharacterConstant(self, o, **kwargs):
        return sym.Literal(value=f'"{o.text}"', type=BasicType.CHARACTER)

    def visit_FintConstant(self, o, **kwargs):
        if 'kind' in o.attrib:
            _type = self.visit(self.type_map[o.attrib.get('type')], **kwargs)
            return sym.Literal(value=int(o.text), type=BasicType.INTEGER, kind=_type.kind)
        return sym.Literal(value=int(o.text), type=BasicType.INTEGER)

    def visit_FcomplexConstant(self, o, **kwargs):
        value = ', '.join(f'{self.visit(v, **kwargs)}' for v in list(o))
        return sym.IntrinsicLiteral(value=f'({value})')

    def visit_FarrayConstructor(self, o, **kwargs):
        values = as_tuple(self.visit(v, **kwargs) for v in o)
        if 'element_type' in o.attrib:
            dtype = self.type_from_type_attrib(o.attrib['element_type'])
        else:
            dtype = None
        return sym.LiteralList(values=values, dtype=dtype)

    def visit_functionCall(self, o, **kwargs):

        if 'is_intrinsic' in o.attrib:
            # Register the ProcedureType in the scope before the name lookup
            pname = o.find('name').text
            proc_type = ProcedureType(
                name=pname, is_function=True, is_intrinsic=True, procedure=None
            )
            kwargs['scope'].symbol_attrs[pname] = SymbolAttributes(dtype=proc_type, is_intrinsic=True)

        if o.find('name') is not None:
            name = self.visit(o.find('name'), **kwargs)
        elif o.find('FmemberRef') is not None:
            name = self.visit(o.find('FmemberRef'), **kwargs)
        else:
            raise ValueError

        args = o.find('arguments')
        if args is not None:
            args = as_tuple(self.visit(a, **kwargs) for a in args)
            # Separate keyword argument from positional arguments
            kw_args = as_tuple(arg for arg in args if isinstance(arg, tuple))
            args = as_tuple(arg for arg in args if not isinstance(arg, tuple))
        else:
            args, kw_args = (), ()

        if o.attrib.get('type', 'Fvoid') == 'Fvoid':
            # Subroutine call
            return ir.CallStatement(name=name, arguments=args, kwarguments=kw_args, source=kwargs['source'])

        if name.name.lower() in ('real', 'int'):
            assert args
            expr = args[0]
            if kw_args:
                assert len(args) == 1
                assert len(kw_args) == 1 and kw_args[0][0] == 'kind'
                kind = kw_args[0][1]
            else:
                kind = args[1] if len(args) > 1 else None
            return sym.Cast(name, expr, kind=kind)

        return sym.InlineCall(name, parameters=args, kw_parameters=kw_args)

    def visit_FstructConstructor(self, o, **kwargs):
        _type = self.type_from_type_attrib(o.attrib['type'], **kwargs)
        assert isinstance(_type.dtype, DerivedType)

        name = sym.Variable(name=_type.dtype.name)
        args = [self.visit(a, **kwargs) for a in o]

        # Separate keyword argument from positional arguments
        kw_args = as_tuple(arg for arg in args if isinstance(arg, tuple))
        args = as_tuple(arg for arg in args if not isinstance(arg, tuple))

        return sym.InlineCall(name, parameters=args, kw_parameters=kw_args)

    def visit_FcycleStatement(self, o, **kwargs):
        # TODO: do-construct-name is not preserved
        return ir.Intrinsic(text='cycle', source=kwargs['source'])

    def visit_continueStatement(self, o, **kwargs):
        return ir.Intrinsic(text='continue', source=kwargs['source'])

    def visit_FexitStatement(self, o, **kwargs):
        # TODO: do-construct-name is not preserved
        return ir.Intrinsic(text='exit', source=kwargs['source'])

    def visit_FopenStatement(self, o, **kwargs):
        nvalues = [self.visit(nv, **kwargs) for nv in o.find('namedValueList')]
        nargs = ', '.join(f'{k}={v}' for k, v in nvalues)
        return ir.Intrinsic(text=f'open({nargs})', source=kwargs['source'])

    def visit_FcloseStatement(self, o, **kwargs):
        nvalues = [self.visit(nv, **kwargs) for nv in o.find('namedValueList')]
        nargs = ', '.join(f'{k}={v}' for k, v in nvalues)
        return ir.Intrinsic(text=f'close({nargs})', source=kwargs['source'])

    def visit_FreadStatement(self, o, **kwargs):
        nvalues = [self.visit(nv, **kwargs) for nv in o.find('namedValueList')]
        values = [self.visit(v, **kwargs) for v in o.find('valueList')]
        nargs = ', '.join(f'{k}={v}' for k, v in nvalues)
        args = ', '.join(f'{v}' for v in values)
        return ir.Intrinsic(text=f'read({nargs}) {args}', source=kwargs['source'])

    def visit_FwriteStatement(self, o, **kwargs):
        nvalues = [self.visit(nv, **kwargs) for nv in o.find('namedValueList')]
        values = [self.visit(v, **kwargs) for v in o.find('valueList')]
        nargs = ', '.join(f'{k}={v}' for k, v in nvalues)
        args = ', '.join(f'{v}' for v in values)
        return ir.Intrinsic(text=f'write({nargs}) {args}', source=kwargs['source'])

    def visit_FprintStatement(self, o, **kwargs):
        values = [self.visit(v, **kwargs) for v in o.find('valueList')]
        args = ', '.join(f'{v}' for v in values)
        fmt = o.attrib['format']
        return ir.Intrinsic(text=f'print {fmt}, {args}', source=kwargs['source'])

    def visit_FformatDecl(self, o, **kwargs):
        fmt = f'FORMAT{o.attrib["format"]}'
        return ir.Intrinsic(text=fmt, source=kwargs['source'])

    def visit_namedValue(self, o, **kwargs):
        name = o.attrib['name']
        if 'value' in o.attrib:
            return name, o.attrib['value']
        return name, self.visit(list(o)[0], **kwargs)

    @staticmethod
    def parenthesize_if_needed(expr, enclosing_cls):
        # Other than FP/OFP, OMNI does not retain any information about parenthesis in the
        # original source. While the parse tree is semantically correct,
        # it may cause problems with some agressively optimising compilers.
        # We inject manual parenthesis here for nested expressions to make sure
        # we capture as much of the evaluation order of the original source as possible.
        # Note: this will result in an abundance of trivial/unnecessary parenthesis!
        if enclosing_cls in (sym.Product, sym.Quotient):
            if isinstance(expr, sym.Product):
                return op.ParenthesisedMul(expr.children)
            if isinstance(expr, sym.Quotient):
                return op.ParenthesisedDiv(expr.numerator, expr.denominator)
            if isinstance(expr, sym.Sum):
                return op.ParenthesisedAdd(expr.children)
            if isinstance(expr, sym.Power):
                return op.ParenthesisedPow(expr.base, expr.exponent)
        return expr

    def visit_plusExpr(self, o, **kwargs):
        exprs = tuple(self.visit(c, **kwargs) for c in o)
        assert len(exprs) == 2
        return sym.Sum(exprs)

    def visit_minusExpr(self, o, **kwargs):
        exprs = tuple(self.visit(c, **kwargs) for c in o)
        assert len(exprs) == 2
        return sym.Sum((exprs[0], sym.Product((-1, exprs[1]))))

    def visit_mulExpr(self, o, **kwargs):
        exprs = tuple(self.visit(c, **kwargs) for c in o)
        assert len(exprs) == 2
        exprs = tuple(self.parenthesize_if_needed(c, sym.Product) for c in exprs)
        return sym.Product(exprs)

    def visit_divExpr(self, o, **kwargs):
        exprs = tuple(self.visit(c, **kwargs) for c in o)
        assert len(exprs) == 2
        exprs = tuple(self.parenthesize_if_needed(c, sym.Quotient) for c in exprs)
        return sym.Quotient(*exprs)

    def visit_FpowerExpr(self, o, **kwargs):
        exprs = tuple(self.visit(c, **kwargs) for c in o)
        assert len(exprs) == 2
        return sym.Power(base=exprs[0], exponent=exprs[1])

    def visit_unaryMinusExpr(self, o, **kwargs):
        exprs = tuple(self.visit(c, **kwargs) for c in o)
        assert len(exprs) == 1
        return sym.Product((-1, exprs[0]))

    def visit_logOrExpr(self, o, **kwargs):
        exprs = tuple(self.visit(c, **kwargs) for c in o)
        return sym.LogicalOr(exprs)

    def visit_logAndExpr(self, o, **kwargs):
        exprs = tuple(self.visit(c, **kwargs) for c in o)
        return sym.LogicalAnd(exprs)

    def visit_logNotExpr(self, o, **kwargs):
        exprs = tuple(self.visit(c, **kwargs) for c in o)
        assert len(exprs) == 1
        return sym.LogicalNot(exprs[0])

    def visit_logLTExpr(self, o, **kwargs):
        exprs = tuple(self.visit(c, **kwargs) for c in o)
        assert len(exprs) == 2
        return sym.Comparison(exprs[0], '<', exprs[1])

    def visit_logLEExpr(self, o, **kwargs):
        exprs = tuple(self.visit(c, **kwargs) for c in o)
        assert len(exprs) == 2
        return sym.Comparison(exprs[0], '<=', exprs[1])

    def visit_logGTExpr(self, o, **kwargs):
        exprs = tuple(self.visit(c, **kwargs) for c in o)
        assert len(exprs) == 2
        return sym.Comparison(exprs[0], '>', exprs[1])

    def visit_logGEExpr(self, o, **kwargs):
        exprs = tuple(self.visit(c, **kwargs) for c in o)
        assert len(exprs) == 2
        return sym.Comparison(exprs[0], '>=', exprs[1])

    def visit_logEQExpr(self, o, **kwargs):
        exprs = tuple(self.visit(c, **kwargs) for c in o)
        assert len(exprs) == 2
        return sym.Comparison(exprs[0], '==', exprs[1])

    def visit_logNEQExpr(self, o, **kwargs):
        exprs = tuple(self.visit(c, **kwargs) for c in o)
        assert len(exprs) == 2
        return sym.Comparison(exprs[0], '!=', exprs[1])

    def visit_logEQVExpr(self, o, **kwargs):
        exprs = tuple(self.visit(c, **kwargs) for c in o)
        assert len(exprs) == 2
        return sym.LogicalOr((sym.LogicalAnd(exprs), sym.LogicalNot(sym.LogicalOr(exprs))))

    def visit_logNEQVExpr(self, o, **kwargs):
        exprs = tuple(self.visit(c, **kwargs) for c in o)
        assert len(exprs) == 2
        return sym.LogicalAnd((sym.LogicalNot(sym.LogicalAnd(exprs)), sym.LogicalOr(exprs)))

    def visit_FconcatExpr(self, o, **kwargs):
        exprs = tuple(self.visit(c, **kwargs) for c in o)
        assert len(exprs) == 2
        return StringConcat(exprs)

    def visit_gotoStatement(self, o, **kwargs):
        label = int(o.attrib['label_name'])
        return ir.Intrinsic(text=f'go to {label: d}', source=kwargs['source'])

    def visit_FstopStatement(self, o, **kwargs):
        code = o.attrib['code']
        return ir.Intrinsic(text=f'stop {code!s}', source=kwargs['source'])

    def visit_statementLabel(self, o, **kwargs):
        return ir.Comment('__STATEMENT_LABEL__', label=o.attrib['label_name'], source=kwargs['source'])

    def visit_FreturnStatement(self, o, **kwargs):
        return ir.Intrinsic(text='return', source=kwargs['source'])



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
from math import gcd as math_gcd
import pytest
import numpy as np

try:
    _ = math_gcd(4,3,2)
    gcd = math_gcd
except TypeError: #Python 3.8 can only handle two arguments
    from functools import reduce
    def gcd(*args):
        return reduce(math_gcd, args)

from loki.analyse.util_linear_algebra import (
    back_substitution,
    generate_row_echelon_form,
    is_independent_system,
    yield_one_d_systems,
)


@pytest.mark.parametrize(
    "upper_triangular_square_matrix, right_hand_side, expected, divison_operation",
    [
        (
            [[2, 1, -1], [0, 0.5, 0.5], [0, 0, -1]],
            [[8], [1], [1]],
            [[2], [3], [-1]],
            lambda x, y: x / y,
        ),
        (
            [[2, 0], [0, 1]],
            [[10], [11]],
            [[5], [11]],
            lambda x, y: x // y,
        ),
    ],
)
def test_backsubstitution(
    upper_triangular_square_matrix, right_hand_side, expected, divison_operation
):
    assert np.allclose(
        back_substitution(
            np.array(upper_triangular_square_matrix),
            np.array(right_hand_side),
            divison_operation,
        ),
        np.array(expected),
    )


@pytest.mark.parametrize(
    "matrix, result",
    [
        ([[2, 0, 1], [0, 2, 0]], [[1, 0, 0.5], [0, 1, 0]]),
        ([[1, -2, 1, 0], [3, 2, 1, 5]], [[1, -2, 1, 0], [0, 1, -0.25, 0.625]]),
        ([[1, -1, -10]], [[1, -1, -10]]),
        ([[0, 0, 0], [0, 0, 0], [0, 0, 0]], [[0, 0, 0], [0, 0, 0], [0, 0, 0]]),
        ([[1, 2, 3], [4, 5, 6], [7, 8, 9]], [[1, 2, 3], [0, 1, 2], [0, 0, 0]]),
        ([[0, 1, 0], [0, 0, 1], [0, 0, 0]], [[0, 1, 0], [0, 0, 1], [0, 0, 0]]),
        (
            [[2, 4, 6, 8], [1, 2, 3, 4], [3, 6, 9, 12]],
            [[1, 2, 3, 4], [0, 0, 0, 0], [0, 0, 0, 0]],
        ),
        ([[0, 0, 0], [1, 0, 2]], [[1, 0, 2], [0, 0, 0]]),
    ],
)
def test_generate_row_echelon_form(matrix, result):
    matrix = np.array(matrix, dtype=float)
    result = np.array(result, dtype=float)

    assert np.allclose(generate_row_echelon_form(matrix), result)


@pytest.mark.parametrize(
    "matrix, result",
    [
        ([[]], [[]]),
        ([[2, 0, 1], [0, 2, 0]], [[1, 0, 0], [0, 1, 0]]),
        ([[1, -2, 1, 0], [3, 2, 1, 5]], [[1, -2, 1, 0], [0, 1, -1, 0]]),
        ([[1, -1, -10]], [[1, -1, -10]]),
        ([[0, 0, 0], [0, 0, 0], [0, 0, 0]], [[0, 0, 0], [0, 0, 0], [0, 0, 0]]),
        ([[1, 2, 3], [4, 5, 6], [7, 8, 9]], [[1, 2, 3], [0, 1, 2], [0, 0, 0]]),
        ([[0, 1, 0], [0, 0, 1], [0, 0, 0]], [[0, 1, 0], [0, 0, 1], [0, 0, 0]]),
        (
            [[2, 4, 6, 8], [1, 2, 3, 4], [3, 6, 9, 12]],
            [[1, 2, 3, 4], [0, 0, 0, 0], [0, 0, 0, 0]],
        ),
    ],
)
def test_enforce_integer_arithmetics_for_row_echelon_form(matrix, result):
    matrix = np.array(matrix, dtype=float)
    result = np.array(result, dtype=float)

    assert np.allclose(
        generate_row_echelon_form(matrix, division_operator=lambda x, y: x // y), result
    )


def _raise_assertion_error(A):
    raise ValueError()


def _require_gcd_condition(A):
    """Check that gcd condition of linear Diophantine equation is satisfied"""
    if A[0, -1] % gcd(*A[0, :-1]) != 0:
        raise ValueError()


@pytest.mark.parametrize(
    "matrix, condition, result",
    [
        ([[1, 2, 3], [4, 5, 6]], _raise_assertion_error, None),
        (
            [[2, 0, 0, -2, -20], [0, 2, -2, 0, -22]],
            _require_gcd_condition,
            [[1, 0, 0, -1, -10], [0, 1, -1, 0, -11]],
        ),
        ([[2, 0, 0, -2, -20], [0, 2, -2, 0, -21]], _require_gcd_condition, None),
    ],
)
def test_require_conditions(matrix, condition, result):
    matrix = np.array(matrix)

    if result is None:
        with pytest.raises(ValueError):
            _ = generate_row_echelon_form(matrix, conditional_check=condition)
    else:
        result = np.array(result)
        assert np.allclose(
            generate_row_echelon_form(matrix, conditional_check=condition), result
        )


@pytest.mark.parametrize(
    "matrix, expected_result",
    [
        (np.array([[1, 0], [0, 1], [0, 0]]), True),
        (np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]), True),
        (np.array([[1, 0, 1], [0, 1, 0], [0, 0, 0]]), False),
        (np.array([[0, 0, 0], [0, 0, 0], [0, 0, 0]]), True),
        (np.array([[1, 0, 0], [0, 0, 1], [0, 1, 0]]), True),
    ],
)
def test_is_independent_system(matrix, expected_result):
    assert is_independent_system(matrix) == expected_result


@pytest.mark.parametrize(
    "matrix, rhs, list_of_lhs_column, list_of_rhs_column",
    [
        (
            np.array([[1, 0], [0, 1], [0, 0]]),
            np.array([[1], [2], [0]]),
            [np.array([[0]]), np.array([[0]]), np.array([[1]]), np.array([[1]])],
            [np.array([[0]]), np.array([[0]]), np.array([[1]]), np.array([[2]])],
        ),
        (
            np.array([[1, 0], [0, 1], [0, 0]]),
            np.array([[1], [2], [1]]),
            [np.array([[0]]), np.array([[0]]), np.array([[1]]), np.array([[1]])],
            [np.array([[1]]), np.array([[1]]), np.array([[1]]), np.array([[2]])],
        ),
        (
            np.array([[0, 0, 0], [0, 0, 0], [0, 0, 0]]),
            np.array([[1], [2], [3]]),
            [np.array([[0], [0], [0]])] * 3,
            [np.array([[1], [2], [3]])] * 3,
        ),
        (  # will even split non independent systems, call is_independent_system before
            np.array([[2, 1], [1, 3]]),
            np.array([[3], [4]]),
            [np.array([[2], [1]]), np.array([[1], [3]])],
            [np.array([[3], [4]]), np.array([[3], [4]])],
        ),
    ],
)
def test_yield_one_d_systems(matrix, rhs, list_of_lhs_column, list_of_rhs_column):
    results = list(yield_one_d_systems(matrix, rhs))
    assert len(results) == len(list_of_lhs_column) == len(list_of_rhs_column)
    for index, (A, b) in enumerate(results):
        assert np.array_equal(A, list_of_lhs_column[index])
        assert np.array_equal(b, list_of_rhs_column[index])



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from loki.batch import Transformation
from loki.expression import symbols as sym
from loki.ir import (
    nodes as ir, FindNodes, Transformer, FindExpressions,
    SubstituteExpressions
)
from loki.tools import as_tuple

from loki.transformations.sanitise import do_resolve_associates
from loki.transformations.utilities import (
    get_integer_variable, get_loop_bounds, check_routine_sequential
)


__all__ = ['SCCBaseTransformation']


class SCCBaseTransformation(Transformation):
    """
    A basic set of utilities used in the SCC transformation. These utilities
    can either be used as a transformation in their own right, or the contained
    class methods can be called directly.

    Parameters
    ----------
    horizontal : :any:`Dimension`
        :any:`Dimension` object describing the variable conventions used in code
        to define the horizontal data dimension and iteration space.
    directive : string or None
        Directives flavour to use for parallelism annotations; either
        ``'openacc'`` or ``None``.
    """

    def __init__(self, horizontal, directive=None):
        self.horizontal = horizontal

        assert directive in [None, 'openacc']
        self.directive = directive

    # TODO: correct "definition" of a pure/elemental routine (take e.g. loki serial into account ...)
    @staticmethod
    def is_elemental(routine):
        """
        Check whether :any:`Subroutine` ``routine`` is an elemental routine.
        Need for distinguishing elemental and non-elemental function to transform
        those in a different way.

        Parameters
        ----------
        routine: :any:`Subroutine`
            The subroutine to check whether elemental
        """
        for prefix in routine.prefix:
            if prefix.lower() == 'elemental':
                return True
        return False

    @classmethod
    def resolve_masked_stmts(cls, routine, loop_variable):
        """
        Resolve :any:`MaskedStatement` (WHERE statement) objects to an
        explicit combination of :any:`Loop` and :any:`Conditional` combination.

        Parameters
        ----------
        routine : :any:`Subroutine`
            The subroutine in which to resolve masked statements
        loop_variable : :any:`Scalar`
            The induction variable for the created loops.
        """
        mapper = {}
        for masked in FindNodes(ir.MaskedStatement).visit(routine.body):
            # TODO: Currently limited to simple, single-clause WHERE stmts
            assert len(masked.conditions) == 1 and len(masked.bodies) == 1
            ranges = [e for e in FindExpressions().visit(masked.conditions[0]) if isinstance(e, sym.RangeIndex)]
            exprmap = {r: loop_variable for r in ranges}
            assert len(ranges) > 0
            assert all(r == ranges[0] for r in ranges)
            bounds = sym.LoopRange((ranges[0].start, ranges[0].stop, ranges[0].step))
            cond = ir.Conditional(condition=masked.conditions[0], body=masked.bodies[0], else_body=masked.default)
            loop = ir.Loop(variable=loop_variable, bounds=bounds, body=(cond,))
            # Substitute the loop ranges with the loop index and add to mapper
            mapper[masked] = SubstituteExpressions(exprmap).visit(loop)

        routine.body = Transformer(mapper).visit(routine.body)

        # if loops have been inserted, check if loop variable is declared
        if mapper and loop_variable not in routine.variables:
            routine.variables += as_tuple(loop_variable)

    @classmethod
    def resolve_vector_dimension(cls, routine, loop_variable, bounds):
        """
        Resolve vector notation for a given dimension only. The dimension
        is defined by a loop variable and the bounds of the given range.

        TODO: Consolidate this with the internal
        `loki.transform.transform_array_indexing.resolve_vector_notation`.

        Parameters
        ----------
        routine : :any:`Subroutine`
            The subroutine in which to resolve vector notation usage.
        loop_variable : :any:`Scalar`
            The induction variable for the created loops.
        bounds : tuple of :any:`Scalar`
            Tuple defining the iteration space of the inserted loops.
        """

        bounds_str = f'{bounds[0]}:{bounds[1]}'

        mapper = {}
        for stmt in FindNodes(ir.Assignment).visit(routine.body):
            ranges = [e for e in FindExpressions().visit(stmt)
                      if isinstance(e, sym.RangeIndex) and e == bounds_str]
            if ranges:
                exprmap = {r: loop_variable for r in ranges}
                loop = ir.Loop(
                    variable=loop_variable, bounds=sym.LoopRange(bounds),
                    body=as_tuple(SubstituteExpressions(exprmap).visit(stmt))
                )
                mapper[stmt] = loop

        routine.body = Transformer(mapper).visit(routine.body)

        # if loops have been inserted, check if loop variable is declared
        if mapper and loop_variable not in routine.variables:
            routine.variables += as_tuple(loop_variable)


    def transform_subroutine(self, routine, **kwargs):
        """
        Apply SCCBase utilities to a :any:`Subroutine`.

        Parameters
        ----------
        routine : :any:`Subroutine`
            Subroutine to apply this transformation to.
        role : string
            Role of the subroutine in the call tree; should be ``"kernel"``
        """
        role = kwargs['role']

        if role == 'kernel':
            self.process_kernel(routine)
        if role == 'driver':
            self.process_driver(routine)

    def process_kernel(self, routine):
        """
        Applies the SCCBase utilities to a "kernel". This consists simply
        of resolving associations, masked statements and vector notation.

        Parameters
        ----------
        routine : :any:`Subroutine`
            Subroutine to apply this transformation to.
        """

        # Bail if routine is marked as sequential or routine has already been processed
        if check_routine_sequential(routine):
            return

        # Bail if routine is elemental
        if self.is_elemental(routine):
            return

        # check for horizontal loop bounds in subroutine symbol table
        bounds = get_loop_bounds(routine, dimension=self.horizontal)

        # Find the iteration index variable for the specified horizontal
        v_index = get_integer_variable(routine, name=self.horizontal.index)

        # Associates at the highest level, so they don't interfere
        # with the sections we need to do for detecting subroutine calls
        do_resolve_associates(routine)

        # Resolve WHERE clauses
        self.resolve_masked_stmts(routine, loop_variable=v_index)

        # Resolve vector notation, eg. VARIABLE(KIDIA:KFDIA)
        self.resolve_vector_dimension(routine, loop_variable=v_index, bounds=bounds)

    def process_driver(self, routine):
        """
        Applies the SCCBase utilities to a "driver". This consists simply
        of resolving associations.

        Parameters
        ----------
        routine : :any:`Subroutine`
            Subroutine to apply this transformation to.
        """

        # Resolve associates, since the PGI compiler cannot deal with
        # implicit derived type component offload by calling device
        # routines.
        do_resolve_associates(routine)



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from functools import partial

from loki.batch import Pipeline

from loki.transformations.hoist_variables import HoistTemporaryArraysAnalysis
from loki.transformations.pool_allocator import TemporariesPoolAllocatorTransformation
from loki.transformations.raw_stack_allocator import TemporariesRawStackTransformation

from loki.transformations.single_column.base import SCCBaseTransformation
from loki.transformations.single_column.annotate import SCCAnnotateTransformation
from loki.transformations.single_column.hoist import SCCHoistTemporaryArraysTransformation
from loki.transformations.single_column.vector import (
    SCCDevectorTransformation, SCCDemoteTransformation, SCCRevectorTransformation
)
from loki.transformations.single_column.vertical import SCCFuseVerticalLoops

__all__ = [
    'SCCVectorPipeline', 'SCCHoistPipeline', 'SCCStackPipeline', 'SCCRawStackPipeline'
]


"""
The basic Single Column Coalesced (SCC) transformation with
vector-level kernel parallelism.

This tranformation will convert kernels with innermost vectorisation
along a common horizontal dimension to a GPU-friendly loop-layout via
loop inversion and local array variable demotion. The resulting kernel
remains "vector-parallel", but with the ``horizontal`` loop as the
outermost iteration dimension (as far as data dependencies
allow). This allows local temporary arrays to be demoted to scalars,
where possible.

The outer "driver" loop over blocks is used as the secondary dimension
of parallelism, where the outher data indexing dimension
(``block_dim``) is resolved in the first call to a "kernel"
routine. This is equivalent to a so-called "gang-vector" parallisation
scheme.

This :any:`Pipeline` applies the following :any:`Transformation`
classes in sequence:
1. :any:`SCCBaseTransformation` - Ensure utility variables and resolve
   problematic code constructs.
2. :any:`SCCDevectorTransformation` - Remove horizontal vector loops.
3. :any:`SCCDemoteTransformation` - Demote local temporary array
   variables where appropriate.
4. :any:`SCCRevectorTransformation` - Re-insert the vecotr loops outermost,
   according to identified vector sections.
5. :any:`SCCAnnotateTransformation` - Annotate loops according to
   programming model (``directive``).

Parameters
----------
horizontal : :any:`Dimension`
    :any:`Dimension` object describing the variable conventions used in code
    to define the horizontal data dimension and iteration space.
block_dim : :any:`Dimension`
    Optional ``Dimension`` object to define the blocking dimension
    to use for hoisted column arrays if hoisting is enabled.
directive : string or None
    Directives flavour to use for parallelism annotations; either
    ``'openacc'`` or ``None``.
trim_vector_sections : bool
    Flag to trigger trimming of extracted vector sections to remove
    nodes that are not assignments involving vector parallel arrays.
demote_local_arrays : bool
    Flag to trigger local array demotion to scalar variables where possible
"""
SCCVectorPipeline = partial(
    Pipeline, classes=(
        SCCFuseVerticalLoops,
        SCCBaseTransformation,
        SCCDevectorTransformation,
        SCCDemoteTransformation,
        SCCRevectorTransformation,
        SCCAnnotateTransformation
    )
)


"""
SCC-style transformation that additionally hoists local temporary
arrays that cannot be demoted to the outer driver call.

For details of the kernel and driver-side transformations, please
refer to :any:`SCCVectorPipeline`

In addition, this pipeline will invoke
:any:`HoistTemporaryArraysAnalysis` and
:any:`SCCHoistTemporaryArraysTransformation` before the final
annotation step to hoist multi-dimensional local temporary array
variables to the "driver" routine, where they will be allocated on
device and passed down as arguments.

Parameters
----------
horizontal : :any:`Dimension`
    :any:`Dimension` object describing the variable conventions used in code
    to define the horizontal data dimension and iteration space.
block_dim : :any:`Dimension`
    Optional ``Dimension`` object to define the blocking dimension
    to use for hoisted column arrays if hoisting is enabled.
directive : string or None
    Directives flavour to use for parallelism annotations; either
    ``'openacc'`` or ``None``.
trim_vector_sections : bool
    Flag to trigger trimming of extracted vector sections to remove
    nodes that are not assignments involving vector parallel arrays.
demote_local_arrays : bool
    Flag to trigger local array demotion to scalar variables where possible
dim_vars: tuple of str, optional
    Variables to be within the dimensions of the arrays to be
    hoisted. If not provided, no checks will be done for the array
    dimensions in :any:`HoistTemporaryArraysAnalysis`.
"""
SCCHoistPipeline = partial(
    Pipeline, classes=(
        SCCFuseVerticalLoops,
        SCCBaseTransformation,
        SCCDevectorTransformation,
        SCCDemoteTransformation,
        SCCRevectorTransformation,
        HoistTemporaryArraysAnalysis,
        SCCHoistTemporaryArraysTransformation,
        SCCAnnotateTransformation
    )
)


"""
SCC-style transformation that additionally pre-allocates a "stack"
pool allocator and associates local arrays with preallocated memory.

For details of the kernel and driver-side transformations, please
refer to :any:`SCCVectorPipeline`

In addition, this pipeline will invoke
:any:`TemporariesPoolAllocatorTransformation` to back the remaining
locally allocated arrays from a "stack" pool allocator that is
pre-allocated in the driver routine and passed down via arguments.

Parameters
----------
horizontal : :any:`Dimension`
    :any:`Dimension` object describing the variable conventions used in code
    to define the horizontal data dimension and iteration space.
block_dim : :any:`Dimension`
    Optional ``Dimension`` object to define the blocking dimension
    to use for hoisted column arrays if hoisting is enabled.
directive : string or None
    Directives flavour to use for parallelism annotations; either
    ``'openacc'`` or ``None``.
trim_vector_sections : bool
    Flag to trigger trimming of extracted vector sections to remove
    nodes that are not assignments involving vector parallel arrays.
demote_local_arrays : bool
    Flag to trigger local array demotion to scalar variables where possible
check_bounds : bool, optional
    Insert bounds-checks in the kernel to make sure the allocated
    stack size is not exceeded (default: `True`)
"""
SCCStackPipeline = partial(
    Pipeline, classes=(
        SCCFuseVerticalLoops,
        SCCBaseTransformation,
        SCCDevectorTransformation,
        SCCDemoteTransformation,
        SCCRevectorTransformation,
        SCCAnnotateTransformation,
        TemporariesPoolAllocatorTransformation
    )
)

"""
SCC-style transformation that additionally pre-allocates a "stack"
pool allocator and replaces local temporaries with indexed sub-arrays
of this preallocated array.

For details of the kernel and driver-side transformations, please
refer to :any:`SCCVectorPipeline`

In addition, this pipeline will invoke
:any:`TemporariesRawStackTransformation` to back the remaining
locally allocated arrays from a "stack" pool allocator that is
pre-allocated in the driver routine and passed down via arguments.

Parameters
----------
horizontal : :any:`Dimension`
    :any:`Dimension` object describing the variable conventions used in code
    to define the horizontal data dimension and iteration space.
block_dim : :any:`Dimension`
    Optional ``Dimension`` object to define the blocking dimension
    to use for hoisted column arrays if hoisting is enabled.
directive : string or None
    Directives flavour to use for parallelism annotations; either
    ``'openacc'`` or ``None``.
trim_vector_sections : bool
    Flag to trigger trimming of extracted vector sections to remove
    nodes that are not assignments involving vector parallel arrays.
demote_local_arrays : bool
    Flag to trigger local array demotion to scalar variables where possible
check_bounds : bool, optional
    Insert bounds-checks in the kernel to make sure the allocated
    stack size is not exceeded (default: `True`)
driver_horizontal : str, optional
    Override string if a separate variable name should be used for the
    horizontal when allocating the stack in the driver.
"""
SCCRawStackPipeline = partial(
    Pipeline, classes=(
        SCCBaseTransformation,
        SCCDevectorTransformation,
        SCCDemoteTransformation,
        SCCRevectorTransformation,
        SCCAnnotateTransformation,
        TemporariesRawStackTransformation
    )
)



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

"""
Collection of utility routines to deal with common array indexing conversions.

"""
from collections import defaultdict
from itertools import count
import operator as op

from loki.batch import Transformation, ProcedureItem
from loki.logging import info
from loki.analyse import dataflow_analysis_attached
from loki.expression import symbols as sym, simplify, symbolic_op, is_constant
from loki.ir import (
    nodes as ir, Assignment, Loop, VariableDeclaration, FindNodes,
    Transformer, FindVariables, SubstituteExpressions, FindInlineCalls
)
from loki.tools import as_tuple, CaseInsensitiveDict
from loki.types import SymbolAttributes, BasicType
from loki.transformations.inline import inline_constant_parameters


__all__ = [
    'remove_explicit_array_dimensions', 'add_explicit_array_dimensions',
    'shift_to_zero_indexing', 'invert_array_indices',
    'resolve_vector_notation', 'normalize_range_indexing',
    'promote_variables', 'promote_nonmatching_variables',
    'promotion_dimensions_from_loop_nest', 'demote_variables',
    'flatten_arrays', 'normalize_array_shape_and_access',
    'LowerConstantArrayIndices'
]

def remove_explicit_array_dimensions(routine, calls_only=False):
    """
    Remove colon notation from array dimensions within :any:`Subroutine` ``routine``.
    E.g., convert two-dimensional array ``arr2d(:,:)`` to ``arr2d`` or
    ``arr3d(:,:,:)`` to ``arr3d``, but NOT e.g., ``arr(1,:,:)``.

    Parameters
    ----------
    routine: :any:`Subroutine`
        The subroutine to check
    calls_only: bool
        Whether to remove colon notation from array dimensions only
        from arrays within (inline) calls or all arrays (default: False)
    """
    if calls_only:
        # handle calls (to subroutines) and inline calls (to functions)
        calls = FindNodes(ir.CallStatement).visit(routine.body)
        inline_calls = FindInlineCalls().visit(routine.body)
        inline_call_map = {}
        for call in as_tuple(calls) + as_tuple(inline_calls):
            # handle arguments
            arguments = ()
            for arg in call.arguments:
                if isinstance(arg, sym.Array) and all(dim == sym.RangeIndex((None, None)) for dim in arg.dimensions):
                    new_dimensions = None
                    arguments += (arg.clone(dimensions=new_dimensions),)
                else:
                    arguments += (arg,)
            # handle kwargs
            kwarguments = ()
            for (kwarg_name, kwarg) in call.kwarguments:
                if isinstance(kwarg, sym.Array) and all(dim==sym.RangeIndex((None, None)) for dim in kwarg.dimensions):
                    kwarguments += ((kwarg_name, kwarg.clone(dimensions=None)),)
                else:
                    kwarguments += ((kwarg_name, kwarg),)
            # distinguish calls and inline calls
            if isinstance(call, sym.InlineCall):
                inline_call_map[call] = call.clone(parameters=arguments, kw_parameters=kwarguments)
            else:
                # directly update calls
                call._update(arguments=arguments, kwarguments=kwarguments)
        if inline_call_map:
            # update inline calls via expression substitution
            routine.body = SubstituteExpressions(inline_call_map).visit(routine.body)
    else:
        arrays = [var for var in FindVariables(unique=False).visit(routine.body) if isinstance(var, sym.Array)]
        array_map = {}
        for array in arrays:
            if all(dim == sym.RangeIndex((None, None)) for dim in array.dimensions):
                new_dimensions = None
                array_map[array] = array.clone(dimensions=new_dimensions)
        routine.body = SubstituteExpressions(array_map).visit(routine.body)


def add_explicit_array_dimensions(routine):
    """
    Make dimensions of arrays explicit within :any:`Subroutine` ``routine``.
    E.g., convert two-dimensional array ``arr2d`` to ``arr2d(:,:)`` or
    ``arr3d`` to ``arr3d(:,:,:)``.

    Parameters
    ----------
    routine: :any:`Subroutine`
        The subroutine to check
    """
    arrays = [var for var in FindVariables(unique=False).visit(routine.body) if isinstance(var, sym.Array)]
    array_map = {}
    for array in arrays:
        if not array.dimensions:
            new_dimensions = (sym.RangeIndex((None, None)),) * len(array.shape)
            array_map[array] = array.clone(dimensions=new_dimensions)
    routine.body = SubstituteExpressions(array_map).visit(routine.body)


def shift_to_zero_indexing(routine, ignore=None):
    """
    Shift all array indices to adjust to 0-based indexing conventions (eg. for C or Python)

    Parameters
    ----------
    routine : :any:`Subroutine`
        The subroutine in which the array dimensions should be shifted
    ignore : list of str
        List of variable names for which, if found in the dimension expression
        of an array subscript, that dimension is not shifted to zero.
    """
    ignore = as_tuple(ignore)
    vmap = {}
    for v in FindVariables(unique=False).visit(routine.body):
        if isinstance(v, sym.Array):
            new_dims = []
            for d in v.dimensions:
                if isinstance(d, sym.RangeIndex):
                    start = d.start - sym.Literal(1) if d.start is not None else None
                    # no shift for stop because Python ranges are [start, stop)
                    new_dims += [sym.RangeIndex((start, d.stop, d.step))]
                else:
                    if ignore and any(var in ignore for var in FindVariables().visit(d)):
                        new_dims += [d]
                    else:
                        new_dims += [d - sym.Literal(1)]
            vmap[v] = v.clone(dimensions=as_tuple(new_dims))
    routine.body = SubstituteExpressions(vmap).visit(routine.body)


def invert_array_indices(routine):
    """
    Invert data/loop accesses from column to row-major

    TODO: Take care of the indexing shift between C and Fortran.
    Basically, we are relying on the CGen to shift the iteration
    indices and dearly hope that nobody uses the index's value.
    """
    # Invert array indices in the routine body
    vmap = {}
    for v in FindVariables(unique=True).visit(routine.body):
        if isinstance(v, sym.Array):
            rdim = as_tuple(reversed(v.dimensions))
            vmap[v] = v.clone(dimensions=rdim)
    routine.body = SubstituteExpressions(vmap).visit(routine.body)

    # Invert variable and argument dimensions for the automatic cast generation
    for v in routine.variables:
        if isinstance(v, sym.Array):
            rdim = as_tuple(reversed(v.dimensions))
            if v.shape:
                rshape = as_tuple(reversed(v.shape))
                vmap[v] = v.clone(dimensions=rdim, type=v.type.clone(shape=rshape))
            else:
                vmap[v] = v.clone(dimensions=rdim)
    routine.variables = [vmap.get(v, v) for v in routine.variables]


def resolve_vector_notation(routine):
    """
    Resolve implicit vector notation by inserting explicit loops
    """
    loop_map = {}
    index_vars = set()
    vmap = {}
    # find available loops and create map {(lower, upper, step): loop_variable}
    loops = FindNodes(Loop).visit(routine.body)
    loop_map = {(loop.bounds.lower, loop.bounds.upper, loop.bounds.step or 1):
                loop.variable for loop in loops}
    for stmt in FindNodes(Assignment).visit(routine.body):
        # Loop over all variables and replace them with loop indices
        vdims = []
        shape_index_map = {}
        index_range_map = {}
        for v in FindVariables(unique=False).visit(stmt):
            if not isinstance(v, sym.Array):
                continue

            # Skip if the entire array is used implicitly
            if not v.dimensions:
                continue

            ivar_basename = f'i_{stmt.lhs.basename}'
            for i, dim, s in zip(count(), v.dimensions, as_tuple(v.shape)):
                if isinstance(dim, sym.RangeIndex):
                    # use the shape for e.g., `ARR(:)`, but use the dimension for e.g., `ARR(2:5)`
                    _s = dim if dim.lower is not None else s
                    # create tuple to test whether an appropriate loop is already available
                    test_range = (sym.IntLiteral(1), _s, 1) if not isinstance(_s, sym.RangeIndex)\
                            else (_s.lower, _s.upper, 1)
                    # actually test for it
                    if test_range in loop_map:
                        # Use index variable of available matching loop
                        ivar = loop_map[test_range]
                    else:
                        # Create new index variable
                        vtype = SymbolAttributes(BasicType.INTEGER)
                        ivar = sym.Variable(name=f'{ivar_basename}_{i}', type=vtype, scope=routine)
                    shape_index_map[(i, s)] = ivar
                    index_range_map[ivar] = _s

                    if ivar not in vdims:
                        vdims.append(ivar)

            # Add index variable to range replacement
            new_dims = as_tuple(shape_index_map.get((i, s), d)
                                for i, d, s in zip(count(), v.dimensions, as_tuple(v.shape)))
            vmap[v] = v.clone(dimensions=new_dims)

        index_vars.update(list(vdims))

        # Recursively build new loop nest over all implicit dims
        if len(vdims) > 0:
            loop = None
            body = stmt
            for ivar in vdims:
                irange = index_range_map[ivar]
                if isinstance(irange, sym.RangeIndex):
                    bounds = sym.LoopRange(irange.children)
                else:
                    bounds = sym.LoopRange((sym.Literal(1), irange, sym.Literal(1)))
                loop = Loop(variable=ivar, body=as_tuple(body), bounds=bounds)
                body = loop

            loop_map[stmt] = loop

    if len(loop_map) > 0:
        routine.body = Transformer(loop_map).visit(routine.body)
    routine.variables += tuple(set(index_vars))

    # Apply variable substitution
    routine.body = SubstituteExpressions(vmap).visit(routine.body)


def normalize_range_indexing(routine):
    """
    Replace the ``(1:size)`` indexing in array sizes that OMNI introduces.
    """
    def is_one_index(dim):
        return isinstance(dim, sym.RangeIndex) and dim.lower == 1 and dim.step is None

    vmap = {}
    for v in routine.variables:
        if isinstance(v, sym.Array):
            new_dims = [d.upper if is_one_index(d) else d for d in v.dimensions]
            new_shape = [d.upper if is_one_index(d) else d for d in v.shape]
            new_type = v.type.clone(shape=as_tuple(new_shape))
            vmap[v] = v.clone(dimensions=as_tuple(new_dims), type=new_type)
    routine.variables = [vmap.get(v, v) for v in routine.variables]


def promote_variables(routine, variable_names, pos, index=None, size=None):
    """
    Promote a list of variables by inserting new array dimensions of given size
    and updating all uses of these variables with a given index expression.

    When providing only :data:`size` or :data:`index`, promotion is restricted
    to updating only variable declarations or their use, respectively, and the
    other is left unchanged.

    Parameters
    ----------
    routine : :any:`Subroutine`
        The subroutine in which the variables should be promoted.
    variable_names : list of str
        The names of variables to be promoted. Matching of variables against
        names is case-insensitive.
    pos : int
        The position of the new array dimension using Python indexing
        convention (i.e., count from 0 and use negative values to count from
        the end).
    index : :py:class:`pymbolic.primitives.Expression`, optional
        The indexing expression (or a tuple for multi-dimension promotion)
        to use for the promotion dimension(s), e.g., loop variables. Usage of
        variables is only updated if `index` is provided. When the index
        expression is not live at the variable use, ``:`` is used instead.
    size : :py:class:`pymbolic.Expression`, optional
        The size of the dimension (or tuple for multi-dimension promotion) to
        insert at `pos`. When this is provided, the declaration of variables
        is updated accordingly.
    """
    variable_names = {name.lower() for name in variable_names}

    if not variable_names:
        return

    # Insert new index dimension
    if index is not None:
        index = as_tuple(index)
        index_vars = [set(FindVariables().visit(i)) for i in index]

        # Create a copy of the tree and apply promotion in-place
        routine.body = Transformer().visit(routine.body)

        with dataflow_analysis_attached(routine):
            for node, var_list in FindVariables(unique=False, with_ir_node=True).visit(routine.body):
                # All the variables marked for promotion that appear in this IR node
                var_list = [v for v in var_list if v.name.lower() in variable_names]

                if not var_list:
                    continue

                # We use the given index expression in this node if all
                # variables therein are defined, otherwise we use `:`
                node_index = tuple(i if v <= node.live_symbols else sym.RangeIndex((None, None))
                                   for i, v in zip(index, index_vars))

                var_map = {}
                for var in var_list:
                    # If the position is given relative to the end we convert it to
                    # a positive index
                    if hasattr(var, 'dimensions'):
                        var_dim = var.dimensions
                    else:
                        var_dim = ()
                    if pos < 0:
                        var_pos = len(var_dim) - pos + 1
                    else:
                        var_pos = pos

                    dimensions = as_tuple(var_dim[:var_pos] + node_index + var_dim[var_pos:])
                    var_map[var] = var.clone(dimensions=dimensions)

                # need to apply update immediately because identical variable use
                # in other nodes might yield same hash but different substitution
                SubstituteExpressions(var_map, inplace=True).visit(node)

    # Apply shape promotion
    if size is not None:
        size = as_tuple(size)

        var_list = [var for decl in FindNodes(VariableDeclaration).visit(routine.spec)
                    for var in decl.symbols if var.name.lower() in variable_names]

        var_shapes = [getattr(var, 'shape', ()) for var in var_list]
        if pos < 0:
            var_pos = [len(shape) - pos + 1 for shape in var_shapes]
        else:
            var_pos = [pos] * len(var_shapes)
        var_shapes = [d[:p] + size + d[p:] for d, p in zip(var_shapes, var_pos)]

        var_map = {v: v.clone(type=v.type.clone(shape=shape), dimensions=shape)
                   for v, shape in zip(var_list, var_shapes)}
        routine.spec = SubstituteExpressions(var_map).visit(routine.spec)


def promotion_dimensions_from_loop_nest(var_names, loops, promotion_vars_dims, promotion_vars_index):
    """
    Determine promotion dimensions corresponding to the iteration space of a loop nest.

    Parameters
    ----------
    var_names : list of str
        The names of variables to consider for promotion.
    loops : list of :any:`Loop`
        The list of nested loops, sorted from outermost to innermost.
    promotion_vars_dims : dict((str, tuple))
        The mapping of variable names to promotion dimensions. When determining
        promotion dimensions for the variables in :data:`var_names` this dict is
        checked for already existing promotion dimensions and, if not matching,
        the maximum of both is taken for each dimension.
    promotion_vars_index : dict((str, tuple))
        The mapping of variable names to subscript expressions. These expressions
        are later inserted for every variable use. When the indexing expression
        for the loop nest does not match the existing expression in this dict,
        a :any:`RuntimeError` is raised.

    Returns
    -------
    (:data:`promotion_vars_dims`, :data:`promotion_vars_dims`) : tuple of dict
        The updated mappings :data:`promotion_vars_dims` and :data:`promotion_vars_index`.

    """
    # TODO: Would be nice to be able to promote this to the smallest possible dimension
    #       (in a loop var=start,end this is (end-start+1) with subscript index (var-start+1))
    #       but it requires being able to decide whether this yields a constant dimension,
    #       thus we need to stick to the upper bound for the moment as this is constant
    #       in our use cases.
    loop_lengths = [simplify(loop.bounds.stop) for loop in reversed(loops)]
    loop_index = [loop.variable for loop in reversed(loops)]

    def _merge_dims_and_index(dims_a, index_a, dims_b, index_b, var_name):
        """
        Helper routine that takes two pairs of promotion dimensions and indices
        (let's call them a and b) and tries to merge them to form the promotion
        configuration that accomodates both.
        """
        # Let's assume we have the same or more promotion dimensions in b than in a
        if len(dims_b) < len(dims_a):
            return _merge_dims_and_index(dims_b, index_b, dims_a, index_a, var_name)  # pylint: disable=arguments-out-of-order

        # We identify each dimension by the index expression; therefore, we have
        # to merge them first
        new_index = []
        ptr_a, ptr_b = 0, 0
        while ptr_a < len(index_a) and ptr_b < len(index_b):
            # Let's see if the next index in a can be found somewhere in b
            try:
                a_in_b = index_b.index(index_a[ptr_a], ptr_b)
            except ValueError:
                a_in_b = None

            if a_in_b is None:
                # It's not in there, so just add it to the new index
                # and go to the next
                new_index += [index_a[ptr_a]]
                ptr_a += 1
            else:
                # Found a in b: add it and anything before from b
                new_index += index_b[ptr_b:a_in_b+1]
                ptr_a += 1
                ptr_b = a_in_b + 1

            # Skip any indices we have already dealt with
            while ptr_a < len(index_a) and index_a[ptr_a] in new_index:
                ptr_a += 1
            while ptr_b < len(index_b) and index_b[ptr_b] in new_index:
                ptr_b += 1

        # Add any remaining indices in a and b
        if ptr_a < len(index_a):
            assert ptr_b == len(index_b)
            new_index += index_a[ptr_a:]
        else:
            assert ptr_a == len(index_a)
            new_index += index_b[ptr_b:]

        # With the merged index in place, we need to go through each corresponding
        # dimension from a and b and pick the larger
        new_dims = []
        for idx in new_index:
            # Look for position of that index in a and b
            try:
                ptr_a = index_a.index(idx)
            except ValueError:
                ptr_a = None
            try:
                ptr_b = index_b.index(idx)
            except ValueError:
                ptr_b = None

            if ptr_a is None:
                # exists only in b
                new_dims += [dims_b[ptr_b]]
            elif ptr_b is None:
                # exists only in a
                new_dims += [dims_a[ptr_a]]
            else:
                # exists in both: pick the larger
                if symbolic_op(dims_a[ptr_a], op.lt, dims_b[ptr_b]):
                    new_dims += [dims_b[ptr_b]]
                else:
                    new_dims += [dims_a[ptr_a]]

        # ... and we're done: return the new dimensions and index
        return new_dims, new_index

    for var_name in var_names:
        # Check if we have already marked this variable for promotion: let's make sure the added
        # dimensions are large enough for this loop (nest)
        if var_name not in promotion_vars_dims:
            promotion_vars_dims[var_name] = loop_lengths
            promotion_vars_index[var_name] = loop_index
        else:
            promotion_vars_dims[var_name], promotion_vars_index[var_name] = \
                _merge_dims_and_index(promotion_vars_dims[var_name], promotion_vars_index[var_name],
                                      loop_lengths, loop_index, var_name)

    return promotion_vars_dims, promotion_vars_index


def promote_nonmatching_variables(routine, promotion_vars_dims, promotion_vars_index):
    """
    Promote multiple variables with potentially non-matching promotion
    dimensions or index expressions.

    This is a convenience routine for using :meth:`promote_variables` that
    groups variables by indexing expression and promotion dimensions to
    reduce the number of calls to :meth:`promote_variables`.

    Parameters
    ----------
    routine : any:`Subroutine`
        The subroutine to be modified.
    promotion_vars_dims : dict
        The mapping of variable names to promotion dimensions. The variables'
        shapes are expanded where necessary to have at least these dimensions.
    promotion_vars_index : dict
        The mapping of variable names to subscript expressions to be used
        whenever reading/writing the variable.
    """
    if not promotion_vars_dims:
        return

    variable_map = routine.variable_map

    # First, let's find out what dimensions we actually need to add
    for var_name in promotion_vars_dims:
        shape = variable_map[var_name].type.shape
        if shape is None:
            continue

        # Eliminate 1:n declared shapes (mostly thanks to OMNI)
        shape = [s.stop if isinstance(s, sym.Range) and s.start == 1 else s for s in shape]

        dims = []
        index = []
        for dim, idx in zip(promotion_vars_dims[var_name], promotion_vars_index[var_name]):
            if not any(symbolic_op(dim, op.eq, d) for d in shape):
                dims += [dim]
                index += [idx]
        promotion_vars_dims[var_name] = dims
        promotion_vars_index[var_name] = index

    # Group promotion variables by index and size to reduce number of traversals for promotion
    index_size_var_map = defaultdict(list)
    for var_name, size in promotion_vars_dims.items():
        index_size_var_map[(as_tuple(promotion_vars_index[var_name]), as_tuple(size))] += [var_name]
    for (index, size), var_names in index_size_var_map.items():
        promote_variables(routine, var_names, -1, index=index, size=size)
    info('%s: promoted variable(s): %s', routine.name, ', '.join(promotion_vars_dims.keys()))


def demote_variables(routine, variable_names, dimensions):
    """
    Demote a list of array variables by removing any occurence of a
    provided set of dimension symbols.

    Parameters
    ----------
    routine : :any:`Subroutine`
        The subroutine in which the variables should be promoted.
    variable_names : list of str
        The names of variables to be promoted. Matching of variables against
        names is case-insensitive.
    dimensions : :py:class:`pymbolic.Expression` or tuple
        Symbol name or tuple of symbol names representing the dimension
        to remove from all occurances of the named variables.
    """
    dimensions = as_tuple(dimensions)

    # Compare lower-case only, since we're not comparing symbols
    vnames = tuple(name.lower() for name in variable_names)

    variables = FindVariables(unique=False).visit(routine.ir)
    variables = tuple(v for v in variables if v.name.lower() in vnames)
    variables = tuple(v for v in variables if hasattr(v, 'shape'))

    if not variables:
        return

    # Record original array shapes
    shape_map = CaseInsensitiveDict({v.name: v.shape for v in variables})

    # Remove shape and dimension entries from each variable in the list
    vmap = {}
    for v in variables:
        old_shape = shape_map[v.name]
        new_shape = tuple(s for s in old_shape if s not in dimensions)
        new_dims = tuple(d for d, s in zip(v.dimensions, old_shape) if s in new_shape)

        new_type = v.type.clone(shape=new_shape or None)
        vmap[v] = v.clone(dimensions=new_dims or None, type=new_type)

    # Propagate the new dimensions to declarations and routine bodys
    routine.body = SubstituteExpressions(vmap).visit(routine.body)
    routine.spec = SubstituteExpressions(vmap).visit(routine.spec)

    # Ensure all declarations with `DIMENSION` keywords are modified too!
    decls = tuple(
        d for d in FindNodes(VariableDeclaration).visit(routine.spec)
        if d.dimensions and any(s.name.lower() in vnames for s in d.symbols)
    )
    decl_map = {}
    for decl in decls:
        # If all symbols have the same shape (after demotion)
        sym_shape = tuple(s.shape if isinstance(s, sym.Array) else None for s in decl.symbols)
        if all(d == sym_shape[0] for d in sym_shape):
            dimensions = decl.symbols[0].shape if isinstance(decl.symbols[0], sym.Array) else None
            decl_map[decl] = decl.clone(dimensions=dimensions)
        else:
            # If not, split into multiple declarations
            sdims = tuple(s.shape if isinstance(s, sym.Array) else None for s in decl.symbols)
            decl_map[decl] = tuple(
                decl.clone(symbols=(s,), dimensions=d) for s, d in zip(decl.symbols, sdims)
            )
    routine.spec = Transformer(decl_map).visit(routine.spec)

    info(f'[Loki::Transform] Demoted variables in {routine.name}: {", ".join(variable_names)}')

def normalize_array_shape_and_access(routine):
    """
    Shift all arrays to start counting at "1"
    """
    def is_explicit_range_index(dim):
        # return False if assumed sized array or lower dimension equals to 1
        # return (isinstance(dim, sym.RangeIndex) and not dim.lower == 1 and not dim is None
        #             and not dim.lower is None and not dim.upper is None)
        return (isinstance(dim, sym.RangeIndex)
                and not (dim.lower == 1 or dim.lower is None or dim.upper is None))

    vmap = {}
    for v in FindVariables(unique=False).visit(routine.body):
        if isinstance(v, sym.Array):
            # skip if e.g., `array(len)`, passed as `call routine(array)`
            if not v.dimensions:
                continue
            new_dims = []
            for i, d in enumerate(v.shape):
                if is_explicit_range_index(d):
                    if isinstance(v.dimensions[i], sym.RangeIndex):
                        start = simplify(v.dimensions[i].start - d.start + 1) if d.start is not None else None
                        stop = simplify(v.dimensions[i].stop - d.start + 1) if d.stop is not None else None
                        new_dims += [sym.RangeIndex((start, stop, d.step))]
                    else:
                        start = simplify(v.dimensions[i] - d.start + 1) if d.start is not None else None
                        new_dims += [start]
                else:
                    new_dims += [v.dimensions[i]]
            if new_dims:
                vmap[v] = v.clone(dimensions=as_tuple(new_dims))
    routine.body = SubstituteExpressions(vmap).visit(routine.body)

    vmap = {}
    for v in routine.variables:
        if isinstance(v, sym.Array):
            new_dims = [sym.RangeIndex((1, simplify(d.upper - d.lower + 1)))
                if is_explicit_range_index(d) else d for d in v.dimensions]
            new_shape = [sym.RangeIndex((1, simplify(d.upper - d.lower + 1)))
                if is_explicit_range_index(d) else d for d in v.shape]
            new_type = v.type.clone(shape=as_tuple(new_shape))
            vmap[v] = v.clone(dimensions=as_tuple(new_dims), type=new_type)
    routine.variables = [vmap.get(v, v) for v in routine.variables]
    normalize_range_indexing(routine)


def flatten_arrays(routine, order='F', start_index=1):
    """
    Flatten arrays, converting multi-dimensional arrays to
    one-dimensional arrays.

    Parameters
    ----------
    routine : :any:`Subroutine`
        The subroutine in which the variables should be promoted.
    order : str
        Assume Fortran (F) vs. C memory/array order.
    start_index : int
        Assume array indexing starts with `start_index`.
    """
    def new_dims(dim, shape):
        if all(_dim == sym.RangeIndex((None, None)) for _dim in dim):
            return None
        if len(dim) > 1:
            if isinstance(shape[-2], sym.RangeIndex):
                raise TypeError(f'Resolve shapes being of type RangeIndex, e.g., "{shape[-2]}" before flattening!')
            _dim = (sym.Sum((dim[-2], sym.Product((shape[-2], dim[-1] - start_index)))),)
            new_dim = dim[:-2]
            new_dim += _dim
            return new_dims(new_dim, shape[:-1])
        return dim

    if order == 'C':
        array_map = {
            var: var.clone(dimensions=new_dims(var.dimensions[::-1], var.shape[::-1]))
            for var in FindVariables().visit(routine.body)
            if isinstance(var, sym.Array) and var.shape and len(var.shape)
        }
    elif order == 'F':
        array_map = {
            var: var.clone(dimensions=new_dims(var.dimensions, var.shape))
            for var in FindVariables().visit(routine.body)
            if isinstance(var, sym.Array) and var.shape and len(var.shape)
        }
    else:
        raise ValueError(f'Unsupported array order "{order}"')

    routine.body = SubstituteExpressions(array_map).visit(routine.body)

    routine.variables = [v.clone(dimensions=as_tuple(sym.Product(v.shape)),
                                 type=v.type.clone(shape=as_tuple(sym.Product(v.shape))))
                         if isinstance(v, sym.Array) else v for v in routine.variables]

class LowerConstantArrayIndices(Transformation):
    """
    A transformation to pass/lower constant array indices down the call tree. 

    For example, the following code:

    .. code-block:: fortran

      subroutine driver(...)
        real, intent(inout) :: var(nlon,nlev,5,nb)
        do ibl=1,10
          call kernel(var(:, :, 1, ibl), var(:, :, 2:5, ibl))
        end do
      end subroutine driver

      subroutine kernel(var1, var2)
        real, intent(inout) :: var1(nlon, nlev)
        real, intent(inout) :: var2(nlon, nlev, 4)
        var1(:, :) = ...
        do jk=1,nlev
          do jl=1,nlon
            var1(jl, jk) = ...
            do jt=1,4
              var2(jl, jk, jt) = ...
            enddo
          enddo
        enddo
      end subroutine kernel

    is transformed to:

    .. code-block:: fortran

      subroutine driver(...)
        real, intent(inout) :: var(nlon,nlev,5,nb)
        do ibl=1,10
          call kernel(var(:, :, :, ibl), var(:, :, :, ibl))
        end do
      end subroutine driver

      subroutine kernel(var1, var2)
        real, intent(inout) :: var1(nlon, nlev, 5)
        real, intent(inout) :: var2(nlon, nlev, 5)
        var1(:, :, 1) = ...
        do jk=1,nlev
          do jl=1,nlon
            var1(jl, jk, 1) = ...
            do jt=1,4
              var2(jl, jk, jt + 2 + -1) = ...
            enddo
          enddo
        enddo
      end subroutine kernel

    Parameters
    ----------
    recurse_to_kernels: bool
        Recurse to kernels, thus lower constant array indices below the driver level for nested
        kernel calls (default: `True`).
    inline_external_only: bool
        Inline only external constant expressions or all of them (default: `False`)
    """

    # This trafo only operates on procedures
    item_filter = (ProcedureItem,)

    def __init__(self, recurse_to_kernels=True, inline_external_only=True):
        self.recurse_to_kernels = recurse_to_kernels
        self.inline_external_only = inline_external_only

    @staticmethod
    def explicit_dimensions(routine):
        """
        Make dimensions of arrays explicit within :any:`Subroutine` ``routine``.
        E.g., convert two-dimensional array ``arr2d`` to ``arr2d(:,:)`` or
        ``arr3d`` to ``arr3d(:,:,:)``.

        Parameters
        ----------
        routine: :any:`Subroutine`
            The subroutine to check
        """
        arrays = [var for var in FindVariables(unique=False).visit(routine.body) if isinstance(var, sym.Array)]
        array_map = {}
        for array in arrays:
            if not array.dimensions:
                new_dimensions = (sym.RangeIndex((None, None)),) * len(array.shape)
                array_map[array] = array.clone(dimensions=new_dimensions)
        routine.body = SubstituteExpressions(array_map).visit(routine.body)

    @staticmethod
    def is_constant_dim(dim):
        """
        Check whether dimension dim is constant, thus, either a constant
        value or a constant range index.

        Parameters
        ----------
        dim: :py:class:`pymbolic.primitives.Expression`
        """
        if is_constant(dim):
            return True
        if isinstance(dim, sym.RangeIndex)\
                and all(child is not None and is_constant(child) for child in dim.children[:-1]):
            return True
        return False

    def transform_subroutine(self, routine, **kwargs):
        role = kwargs['role']
        targets = tuple(str(t).lower() for t in as_tuple(kwargs.get('targets', None)))
        if role == 'driver' or self.recurse_to_kernels:
            inline_constant_parameters(routine, external_only=self.inline_external_only)
            self.process(routine, targets)

    def process(self, routine, targets):
        """
        Process the driver and possibly kernels
        """
        dispatched_routines = ()
        offset_map = {}
        for call in FindNodes(ir.CallStatement).visit(routine.body):
            if str(call.name).lower() not in targets:
                continue
            # skip already dispatched routines but still update the call signature
            if call.routine in dispatched_routines:
                self.update_call_signature(call)
                continue
            # explicit array dimensions for the callee
            self.explicit_dimensions(call.routine)
            dispatched_routines += (call.routine,)
            # create the offset map and apply to call and callee
            offset_map[call.routine.name.lower()] = self.create_offset_map(call)
            self.process_callee(call.routine, offset_map[call.routine.name.lower()])
            self.update_call_signature(call)

    def update_call_signature(self, call):
        """
        Replace constant indices for call arguments being arrays with ':' and update the call.
        """
        new_args = [arg.clone(dimensions=\
                tuple(sym.RangeIndex((None, None)) if self.is_constant_dim(d) else d for d in arg.dimensions))\
                if isinstance(arg, sym.Array) else arg for arg in call.arguments]
        new_kwargs = [(kw[0], kw[1].clone(dimensions=\
                tuple(sym.RangeIndex((None, None)) if self.is_constant_dim(d) else d for d in kw[1].dimensions)))\
                if isinstance(kw[1], sym.Array) else kw for kw in call.kwarguments]
        call._update(arguments=as_tuple(new_args), kwarguments=as_tuple(new_kwargs))

    def create_offset_map(self, call):
        """
        Create map/dictionary for arguments with constant array indices.
        
        For, e.g., 

        integer :: arg(len1, len2, len3, len4)
        call kernel(..., arg(:, 2, 4:6, i), ...)

        offset_map[arg] = {
            0: (0, None, None),  # same index as before, no offset
            1: (None, 1, len2),  # New index, offset 1, size of the dimension is len2
            2: (1, 4, len3),     # Used to be position 1, offset 4, size of the dimension is len3
            3: (-1, None, None), # disregard as this is neither constant nor passed to callee
        }
        """
        offset_map = {}
        for routine_arg, call_arg in call.arg_iter():
            if not isinstance(routine_arg, sym.Array):
                continue
            offset_map[routine_arg.name] = {}
            current_index = 0
            for i, dim in enumerate(call_arg.dimensions):
                if self.is_constant_dim(dim):
                    if isinstance(dim, sym.RangeIndex):
                        # constant array index is e.g. '1:3' or '5:10'
                        offset_map[routine_arg.name][i] = (current_index, dim.children[0], call_arg.shape[i])
                    else:
                        # constant array index is e.g., '1' or '42'
                        offset_map[routine_arg.name][i] = (None, dim, call_arg.shape[i])
                        current_index -= 1
                else:
                    if not isinstance(dim, sym.RangeIndex):
                        # non constant array index is a variable e.g. 'jl'
                        offset_map[routine_arg.name][i] = (-1, None, None)
                        current_index -= 1
                    else:
                        # non constant array index is ':'
                        offset_map[routine_arg.name][i] = (current_index, None, None)
                current_index += 1
        return offset_map

    def process_callee(self, routine, offset_map):
        """
        Process/adapt the callee according to information in `offset_map`.

        Adapt the variable declarations and usage/indexing.
        """
        # adapt variable declarations, thus adapt the dimension and shape of the corresponding arguments
        vmap = {}
        variable_map = routine.variable_map
        for var_name in offset_map:
            var = variable_map[var_name]
            new_dims = ()
            for i in range(max(k for k, v in offset_map[var.name].items() if v != 0) + 1):
                original_index = offset_map[var_name][i][0]
                offset = offset_map[var_name][i][1]
                size = offset_map[var_name][i][2]
                if not (original_index is None or 0 <= original_index < len(var.dimensions)):
                    continue
                if offset is not None:
                    new_dims += (size,)
                else:
                    new_dims += (var.shape[original_index],)
            vmap[var] = var.clone(dimensions=new_dims, type=var.type.clone(shape=new_dims))
        routine.spec = SubstituteExpressions(vmap).visit(routine.spec)
        # adapt the variable usage, thus the indexing/dimension
        vmap = {}
        for var in FindVariables(unique=False).visit(routine.body):
            if var.name in offset_map and var.dimensions is not None and var.dimensions:
                new_dims = ()
                for i in range(max(k for k, v in offset_map[var.name].items() if v != 0) + 1):
                    original_index = offset_map[var.name][i][0]
                    offset = offset_map[var.name][i][1]
                    if not (original_index is None or 0 <= original_index < len(var.dimensions)):
                        continue
                    if offset is not None:
                        if original_index is None:
                            new_dims += (offset,)
                        else:
                            new_dims += (var.dimensions[original_index] + offset - 1,)
                    else:
                        new_dims += (var.dimensions[original_index],)
                vmap[var] = var.clone(dimensions=new_dims)
        routine.body = SubstituteExpressions(vmap).visit(routine.body)
